

/*
 *  Filter_adaptive.c	Spatial domain locally adaptive filters
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "filter_adaptive.h"
#include "filter.h"
#include "application.h"
#include "imgbuf.h"
#include "functions.h"
#include "functions2.h"
#include "gtkhelpers.h"
#include "statop.h"
#include "viz.h"






/* Locally adaptive spatial-domain lowpass filter,
   also called the adaptive minimum mean-squared error (AMMSE) filter.

   We compute the local variance in a NxN neighborhood and
   assume that the minimum value represents the noise variance
   (noise and no other features). Then we use the quotient
   g = minimum local variance / local variance as a metric
   of the image "featureness". For g->1 we perform strong
   averaging, and for g->0 we leave the original value.

	NOTE: This filter in its original form (by J.S. Lee, 1980) is terribly weak
	as the smallest local variance is usually a very small number. Some measures have
	been taken to make the filter action stronger and the filter more robust in general:
		(1) We use a Gaussian convolution to generate the filtered component instead of box averaging
		(2) We use Knuth's formula for standard deviation instead of the simple "pocket calculator formula"
		(3) Minimum local variance is determined as the 5% histogram quantile rather than the minimum
			over the entire image. In this fashion, the minimum variance is less dependent on a small
			random region, and the estimated noise (minimum variance) is more robust.
		(4) we intentionally use stdev rather than variance since it strengthens the filter.
		(5) plus, there is an adjustable gamma, in which the coefficient can be made to "bulge"
			further in the direction of g -> 1. Note that this is on top of the gamma=2 that we
			implicitly introduce when we use the standard deviation.
		(6) To compensate for the use of a Gaussian kernel rather than box averaging,
			the blurring kernel size is made 2 pixels wider than the AMMSE size.

	TODO:	(7) To be tested: If the local variance is computed from the difference of original - blurred image?
				The difference contains, after all, more dominantly the noise component.
			(8) ... and trying Donoho's wavelet-based noise estimator??

	Clearly, with some of the measures taken above (notably the gamma correction and the use
	of the stdev), Lee's original statistics that lead to the claim of a *minimum squared error*
	are violated. However, the assumed squared-error optimality is doubtful anyway.

	For a feeling about the quantile: It is fixed to 5%. This means that in a 512x512 image,
	approximately 13,000 of the 262,000 pixels have a lower local variance (and for those,
	g is clamped to 1.0 and the fully blurred image is taken).

	TODO: RGB is uglier than need be - we should have a union type that allows
			to access the RGB elements as array (a job for defs.h).

*/



/* This helper function computes the adaptive filter gain g.
	The filtered image F is computed from the original image I and the 
	blurred image B through

			F = I - g (I-B) = (1-g) I + g B

	Therefore, if sd is large, g->0 and F is approximately the same as I
	(this is assumed to happen in the presence of edges or other strong
	image features that need to be preserved)
	Conversely, if sd is near minsd -- the assumed sd of the noise component --
	we have g -> 1, and F is approximately B. 

	This step can be as simple as g = minsd/sd, but
	- we trap cases of zero divided by zero
	- we clamp the filter to 1 in case sd > minsd (when a quantile is used)
	- we allow strengthening with a form of gamma correction. Typically, sd >> minsd,
	  and g tends to favor the unblurred image. With a gamma correction,
	  g <- g ^ (1/gamma), gamma >= 1, the curve "bulges" to keep closer to 1
	  for a wider range of low sd values.

	NB. If something is out of range, g defaults to 0, and blur is kept to a minimum.
*/

float ammse_gain (float minsd, float sd, float gamma)
{
float g;

	if ( F_ISZERO(minsd) )
		return 0.0;

	if (sd <= minsd)		/* Clamp to 1 for small sd values */
		return 1.0;			/* Relevant when the quantile is used for minsd selection */

	g = minsd / sd;			/* So this should never be > 1 */
	if (gamma > 1.0)
		g = pow (g, 1.0/gamma);

	return g;
}



void local_adaptive_lowpass (image_type* img, int nghsize, float gamma)
{
image_type locvar, blurred;
double buf, avg[3], sd[3], minsd[3], g;
float min, max;
float* p;
int cnt;
int slice,x,y,i,j,k,x1,y1,e;
rgbtriplet rgb, blurred_rgb;
int sdsize;
double *sddata;


	/* Allocate two one-slice images: One for the local variance, and one
		for the Gaussian-blurred version. Then apply Gaussian blurring. */

	e=1; if (img->imgtype==B_RGB) e=3;
	if (allocate_image (&locvar, B_FLOAT, img->xmax, img->ymax, 2*e)==-1)
		fatal_memerr ();
	p = locvar.data;

	if (allocate_image (&blurred, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();

	/* We'll use Knuth's standard deviation formula, which requires the original data
		in a linear array. This array gets allocated here. Intentionally a few elements 
		larger than needed (and x3 for RGB). */

	sdsize = SQR(nghsize*2+3);
	sddata = calloc ( 3*sdsize , sizeof (double));

	/* The following code performs the filter action slice-by-slice */

	for (slice=0; slice < img->zmax; slice++)
	{

		copybuf_z (img, &blurred, slice, slice);		/* Copy slice data */
		gaussian_general (&blurred, nghsize+1, 0);		/* Apply non-adaptive Gaussian blurring, */
														/* One px wider than specified to compensate for the Gaussian */

		/* Pass 1: Find each pixel's local variance */
			
		minsd[0]=minsd[1]=minsd[2]=999999999;
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				for (k=0; k<3; k++)
				{
					avg[k]=0.0; sd[k]=0.0;
				}

				/* Collect neighborhood data points */

				cnt=0; 
				for (j=-nghsize; j<=nghsize; j++)
					for (i=-nghsize; i<=nghsize; i++)
					{
						x1=x+i; y1=y+j;
						if ((x1>=0) && (x1<img->xmax) && (y1>=0) && (y1<img->ymax))
						{
							if (img->imgtype==B_RGB)
							{
								rgb = readbuf_rgb (img,x1,y1,slice);
								sddata[cnt           ] = rgb.red;
								sddata[cnt + sdsize  ] = rgb.green;
								sddata[cnt + 2*sdsize] = rgb.blue;
							}
							else
								sddata[cnt] = readbuf_flt (img,x1,y1,slice);

							cnt++;
						}
					}

				/* From those data points, compute the standard deviation.
					The SD function automatically drops out with a return value
					of zero if there are not enough data points, although this is safe --
					the minimum (3x3 window, corner pixel) is N=4  */

				if (img->imgtype==B_RGB)
				{
					for (k=0; k<3; k++)
					{
						avg[k] = knuths_mean (&(sddata[k*sdsize]), cnt);
						sd[k] = knuths_sdev (&(sddata[k*sdsize]), cnt);
						/*sd /= fabs(avg);			/* ... or local coefficient of variation */
	
						p[x + locvar.xmax*(y + k*locvar.ymax) ] = sd[k];
						p[x + locvar.xmax*(y + (k+3)*locvar.ymax)] = avg[k];
						if ((sd[k]>0) && (avg[k] > 0) && (x>nghsize) && (y>nghsize) && (x<locvar.xmax-nghsize) && (y<locvar.ymax-nghsize))
						{
							if (minsd[k] > sd[k]) minsd[k] = sd[k];
						}
					}
				}
				else
				{
					avg[0] = knuths_mean (sddata, cnt);
					sd[0] = knuths_sdev (sddata, cnt);
					/*sd /= fabs(avg);			/* ... or local coefficient of variation */

					p[x + locvar.xmax*y ] = sd[0];
					p[x + locvar.xmax*(y +locvar.ymax)] = avg[0];
					if ((sd[0]>0) && (avg[0] > 0) && (x>nghsize) && (y>nghsize) && (x<locvar.xmax-nghsize) && (y<locvar.ymax-nghsize))
					{
						if (minsd[0] > sd[0]) minsd[0] = sd[0];
					}
				}

			}

		/* End of pass 1 -- collect local standard deviation */

		if (img->imgtype==B_RGB)	
			dp (2, "AMMSE: Minimum variance is RGB(%f, %f, %f)\n",minsd[0],minsd[1],minsd[2]);
		else
			dp (2, "AMMSE: Minimum variance is %f\n",minsd[0]);


		/* Adjust the local variance image's min & max for cleanliness. Note that these go to fmin, fmax, always. */
		
		if (img->imgtype == B_RGB)
		{
			min = max = p[0];
			for (x=0; x<e*locvar.xmax*locvar.ymax; x++)
			{
				if (min > p[x]) min = p[x];
				if (max < p[x]) max = p[x];
			}
			locvar.minval.fmin = min;
			locvar.maxval.fmax = max;

		}
		else
		{
			min = max = p[0];
			for (x=0; x<locvar.xmax*locvar.ymax; x++)
			{
				if (min > p[x]) min = p[x];
				if (max < p[x]) max = p[x];
			}
			locvar.minval.fmin = min;
			locvar.maxval.fmax = max;
		}

		#ifndef no_use_q5

		if (img->imgtype == B_RGB)		/* Check if we can use a quantile for minimum variance */
		{
			minsd[0] = find_quantile_standalone (&locvar, 0, 256, 5.0, 0);	/* 5% quantile, red channel */
			minsd[1] = find_quantile_standalone (&locvar, 1, 256, 5.0, 0);	/* 5% quantile, green channel */
			minsd[2] = find_quantile_standalone (&locvar, 2, 256, 5.0, 0);	/* 5% quantile, blue channel */
		}
		else
			minsd[0] = find_quantile_standalone (&locvar, 0, 256, 5.0, 0);	/* Use the 5% quantile of the variance distribution */

		if (img->imgtype==B_RGB)	
			dp (1, "AMMSE: Q5 Minimum variance is (%f, %f, %f)\n",minsd[0],minsd[1],minsd[2]);
		else
			dp (1, "AMMSE: Q5 Minimum variance is %f\n",minsd[0]);

		#endif

		/* Pass 2: Adaptive local average.
			Note: For debugging, the next section can be enabled to return the weight g.
			We should see g -> 100% in flat, noisy regions and g->0 in the presence of strong
			image features. This is a debugging function and has limited scope. */

// 		for (y=0; y<img->ymax; y++)
// 			for (x=0; x<img->xmax; x++)
// 			{
// 				buf = p[x + locvar.xmax*y ];
// 				g = ammse_gain (minsd[0], buf, gamma);		/* Filter coefficient only */
// 				p[x + locvar.xmax*y ] = 100.0*g;			/* Scale up to obtain percent */
// 			}
// 		handback (img, &locvar);							/* Return the image of weight coefficients */
// 															/* This crashes out of the z-loop and more, but,
// 																it is not intended for regular use anyway. */
// 		return;

		/* Real Pass 2 begins here */


		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (img->imgtype!=B_RGB)
				{
					buf = readbuf_flt (img,x,y,slice);				/* Original image pixel */
					avg[0] = readbuf_flt (&blurred, x, y, 0);		/* Gaussian-convolved version. */
					sd[0] = p[x + locvar.xmax*y ];					/* Local std.dev for g coefficient */

					g = ammse_gain (minsd[0], sd[0], gamma);

					buf = buf - g * (buf-avg[0]);						/* Adaptive filter formulation is here, actually */
				}

				switch (img->imgtype)		/* Clamps if needed, and stores in original image data format */
				{
					case B_CHAR:
					{
						unsigned char* q = img->data;
						int c = IROUND(buf);
						c = CLAMP (c,0,255);
						q[x + img->xmax*(y + img->ymax*slice)]=c;
						break;
					}
					case B_SHORT:
					{
						short* q = img->data;
						int s = IROUND(buf);
						q[x + img->xmax*(y + img->ymax*slice)]=s;
						break;
					}
					case B_FLOAT:
					{
						float* q = img->data;
						q[x + img->xmax*(y + img->ymax*slice)]=buf;
						break;
					}
					case B_RGB:
					{
						rgbtriplet* q = img->data;
						int c;
						rgb = readbuf_rgb (img,x,y,slice);
						blurred_rgb = readbuf_rgb (&blurred, x,y,0);
						for (k=0; k<3; k++)
						{
							sd[k] = p[x + locvar.xmax*(y + k*locvar.ymax) ];
							avg[k] = p[x + locvar.xmax*(y + (k+3)*locvar.ymax)];
							g = ammse_gain (minsd[k], sd[k], gamma);

							if (k==0)
							{
								c = IROUND(rgb.red - g * (rgb.red-blurred_rgb.red));
								c = CLAMP (c,0,255);
								rgb.red = c;
							}
							else if (k==1)
							{
								c = IROUND(rgb.green - g * (rgb.green-blurred_rgb.green));
								c = CLAMP (c,0,255);
								rgb.green = c;
							}
							else 
							{
								c = IROUND(rgb.blue - g * (rgb.blue-blurred_rgb.blue));
								c = CLAMP (c,0,255);
								rgb.blue = c;
							}
						}
						q[x + img->xmax*(y + img->ymax*slice)]=rgb;
						break;
					}
				}
			}

	}				/* End for all slices ... */

	freebuf (&locvar);			/* Intermediate image with mean and variance no longer needed */
	freebuf (&blurred);
	free (sddata);
	img->uid = next_uid++;
}



/* The driver for the above function for the GUI and with the main image */

void local_adaptive_lowpass_main (int nghsize, float gamma)
{
int z;

	local_adaptive_lowpass (&mainimg, nghsize, gamma);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
}




/* Three-state Median filter,
	Chen, Ma & Chen, IEEE Trans Image Proc 1999; 8(12): 1834-8

	A nonlinear filter for shot-noise removal. Consider two filters,
		1- the regular Rank filter
		2- a rank filter with repetition of the central element (=center weighted median filter)
	The CWM filter is weaker, but has better edge preservation. The rank filter
	is stronger. The tristate median filter is decision-based. For each pixel,
	we compute
		a) the distance d1 of the CWMF to the original pixel
		b) the distance d2 of the rank filter to the original pixel.
	It can be shown that d2>d1, so given a threshold T, we decide as follows:
		if T > d1, do not change the pixel
		if d2 < T < d1, substitute pixel by CWMF output
		if T < d2, substitute pixel by rank filter output

	NOTE:
	In this definition, the three-state median filter operates on 3x3 neighborhoods only.
	Furthermore, the median filter is 2D, operating slice by slice.
	In the CWM definition, we use a fixed repetition factor of 4

	TODO: RGB is handled in the usual ugly way, forcing conversion to grayscale. Yuck.
	TODO: Would there be a sophisticated way to guess a threshold from the d1 and d2 images?

	What about the treshold, anyway? The extreme, T==0, leads to the simple rank filter.
	A very large threshold leaves the image unmodified.  Threshold values in-between
	yield the adaptive behavior, and a smaller threshold leads to a stronger filter action. 
	This needs to be empirically studied with a training set
	where the mean-squared error (or any other distance) is minimized.

	TODO: Something is wrong. With T=0 we should have rank filter behavior, but we don't.

*/


void tristate_median_filter (image_type *img, int z, float threshold)
{
image_type cwmf_img;		/* One slice to store the CWM filter data */
image_type rank_img;		/* One lisce to store the rank data */
float *pc; float *pr;		/* Pointers to cwmf and rank data */
int REPT=4;					/* CWMF repetition factor MUST be an even number! */
int N=1;					/* Neighborhood size. Keep value 1 or need to adjust array sizes */
float median1[9], median2[9+REPT];
float buf,d1,d2;
int i,j,k1,k2,x,y,n,idx;




	if (allocate_image (&cwmf_img, B_FLOAT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
	pc = cwmf_img.data;
	if (allocate_image (&rank_img, B_FLOAT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
	pr = rank_img.data;

	/* PASS 1: Generate the rank and CWMF results */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{

			k1=0; k2=0;
			for (j=-N; j<=N; j++)
				for (i=-N; i<=N; i++)
				{
					buf = readbuf_flt (img,x+i,y+j,z);
					median1[k1++]=buf; median2[k2++]=buf;
					if ((i==0) && (j==0))
					{
						for (n=1; n<REPT; n++)
							median2[k2++]=buf;			/* center-weighted */
					}
				}

			combsort_f (median1, NULL, 9);
			combsort_f (median2, NULL, 9+REPT);

			idx = x + img->xmax * y;
			pr[idx] = median1[4];					/* Rank filter */
			pc[idx] = median2[ (9+REPT)/2 ];		/* cwmf. This is why REPT must be EVEN */
		}

	/* PASS 2: Perform adaptive three-state filtering */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			idx = x + img->xmax * y;
			buf = readbuf_flt (img,x,y,z);
			d1 = fabs (buf - pr[idx]);			/* Difference to rank filter */
			d2 = fabs (buf - pc[idx]);			/* Difference to cwm filter */
			if ((threshold < d1) && (threshold >= d2))
				buf = pc[idx];
			else if (threshold < d2)
				buf = pr[idx];
			/* else our buffer remains unchanged */

			switch (img->imgtype)
			{
				case B_CHAR:
				{
					unsigned char* q = img->data;
					int c = (int) (0.5+buf);
					if (c<0) c=0; else if (c>255) c=255;
					q[x + img->xmax*(y + img->ymax*z)]=c;
					break;
				}
				case B_SHORT:
				{
					short* q = img->data;
					int s = (int) (0.5+buf);
					q[x + img->xmax*(y + img->ymax*z)]=s;
					break;
				}
				case B_FLOAT:
				{
					float* q = img->data;
					q[x + img->xmax*(y + img->ymax*z)]=buf;
					break;
				}
				case B_RGB:
				{
					rgbtriplet* q = img->data;
					int c = (int) (0.5+buf);
					if (c<0) c=0; else if (c>255) c=255;
					q[x + img->xmax*(y + img->ymax*z)]=(rgbtriplet){c,c,c};
					break;
				}
			}
		}

	img->uid = next_uid++;


}


void tristate_median_macro (image_type *img, float thresh)
{
int z;

	for (z=0; z<img->zmax; z++)
		tristate_median_filter (img, z, thresh);
}


void tristate_median_main (float thresh)
{
int z;

	for (z=0; z<mainimg.zmax; z++)
		tristate_median_filter (&mainimg, z, thresh);
		
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
}



/* 	*********  Anisotropic Diffusion Filter ***************

	An improved version of Gaussian blur which retains (or even sharpens)
	edges while reducing noise.
	Perona P, Malik J. Scale-Space and Edge Detection Using Anisotropic
	Diffusion. IEEE Trans Pattern Anal Mach Intell 1990; 12: 629-39.

	Works on 2D on a slice-by-slice basis

*/


/* The "conduction" function. Must be a nonnegative, monotonically falling
	function with a normalization of g(0)=1.
	Variations are possible. Suggestions
	0) g(grad I) = constant 1 leads to Gaussian blurring
	1) g(grad I) = exp (- (|grad I| / K)^2 ) emphasizes high-contrast edges over low-contrast ones
	2) g(grad I) = 1/(1 + ( |grad I| / K)^2) emphasizes wide regions over smaller ones
*/

float anisotropic_diffusion_g (float gradi, float k)
{
float f;

	f = exp (- SQR(gradi/k));
	return f;
}


/* This is the core function. It iterates iter time steps of anisotropic diffusion 
	Lambda is the diffusion strength and should be 0< lambda < 0.25
	k is related to the size of discontinuities. A larger k gives a stronger filter.

	TODO: RGB is handled in the usual ugly way, forcing conversion to grayscale.
*/

void anisotropic_diffusion_iter (image_type *img, float k, float lambda, int iter)
{
image_type target1, target2;
int x,y,z, idx, itercnt;
char ngh8;								/* Internally select neighborhood */
float buf, cn,cw,ce,cs, in,iw,is,ie;
float cnw, cne, csw, cse, inw, ine, isw, ise;	/* Diagonal elements */
float *p1, *p2;
float deltat;							/* Discrete time step */


	if (allocate_image (&target1, B_FLOAT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
	if (allocate_image (&target2, B_FLOAT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
	p1 = target1.data;
	p2 = target2.data;

	ngh8=1;				/* Enable 8-neighborhood */
	if (ngh8)
		deltat = lambda/1.71;		/* Adjustment for the diagonal neighbor's contribution */
	else
		deltat = lambda;


	/* We implement iterations here, operating on a slice-by-slice basis with an intermediate-storage slice */

	for (z=0; z<img->zmax; z++)
	{
		dp (2, "Anisotropic diffusion: Slice %d\n",z);

		/* Get one slice worth of data */

		idx = 0;
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				p1[idx++] = readbuf_flt (img,x,y,z);


		for (itercnt=0; itercnt < iter; itercnt++)
		{

			for (y=0; y<img->ymax; y++)
			{
				idx = img->xmax*y;
				for (x=0; x<img->xmax; x++)
				{
					buf = p1[idx];
	
					/* Compute the four/eight intensity gradients (note sign difference to paper */
	
					in = iw = is = ie = 0.0; inw = ine = isw = ise = 0.0;

					if (y>0) in = buf - p1[idx-img->xmax];
					if (x>0) iw = buf - p1[idx-1];
					if (y<img->ymax-1) is = buf - p1[idx+img->xmax];
					if (x<img->xmax-1) ie = buf - p1[idx+1];
					if (ngh8)
					{
						if ((y>0) && (x>0)) inw = buf - p1[idx - 1 -img->xmax];
						if ((y<img->xmax-1) && (x>0)) isw = buf - p1[idx - 1 + img->xmax];
						if ((y<img->ymax-1) && (x<img->xmax-1)) ise = buf - p1[idx + 1 + img->xmax];
						if ((y>0) && (x<img->xmax-1)) ine = buf - p1[idx + 1 - img->xmax];
					}
	
					/* Compute the four conduction factors - those reaching out of the image are zero */
	
					cn = cw = ce = cs = 0.0; cnw = cne = cse = csw = 0.0;
	
					if (y>0) cn = anisotropic_diffusion_g (in,k);
					if (x>0) cw = anisotropic_diffusion_g (iw,k);
					if (y<img->ymax-1) cs = anisotropic_diffusion_g (is,k);
					if (x<img->xmax-1) ce = anisotropic_diffusion_g (ie,k);
					if (ngh8)
					{
						if ((y>0) && (x>0)) cnw = anisotropic_diffusion_g (inw,k);
						if ((y<img->xmax-1) && (x>0)) csw = anisotropic_diffusion_g (isw,k);
						if ((y<img->ymax-1) && (x<img->xmax-1)) cse = anisotropic_diffusion_g (ise,k);
						if ((y>0) && (x<img->xmax-1)) cne = anisotropic_diffusion_g (ine,k);
					}

					/* Compute the new iteration value at (x,y) and store in target.
							Note sign difference to paper */
	
					buf = buf - deltat*(cn*in + cw*iw + cs*is + ce*ie);
					if (ngh8) buf = buf - 0.71*deltat*(cnw*inw + cne*ine + csw*isw + cse*ise);
					p2[idx] = buf;
					idx++;
				}
			}

			/* Copy back into source buffer. A bit inefficient (we should swap buffer functions)
				but easily implemented and debugged */

			for (idx=0; idx<img->xmax * img->ymax; idx++)
				p1[idx] = p2[idx];
		}

		/* Iterations finished. Now store slice data back in source image */

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = p1[x + img->xmax*y];
				switch (img->imgtype)
				{
					case B_CHAR:
					{
						unsigned char* q = img->data;
						int c = (int) (0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						q[x + img->xmax*(y + img->ymax*z)]=c;
						break;
					}
					case B_SHORT:
					{
						short* q = img->data;
						int s = (int) (0.5+buf);
						q[x + img->xmax*(y + img->ymax*z)]=s;
						break;
					}
					case B_FLOAT:
					{
						float* q = img->data;
						q[x + img->xmax*(y + img->ymax*z)]=buf;
						break;
					}
					case B_RGB:
					{
						rgbtriplet* q = img->data;
						int c = (int) (0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						q[x + img->xmax*(y + img->ymax*z)]=(rgbtriplet){c,c,c};
						break;
					}
				}
			}

	}

	freebuf (&target1);
	freebuf (&target2);
	make_minmax (img);
	img->uid = next_uid++;

}



void anisotropic_diffusion_main (float thresh, int iter)
{

	anisotropic_diffusion_iter (&mainimg, thresh, 0.2, iter);
		
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
}



/* 	*********  Adaptive Bilateral Filter ***************

	A Gaussian filter that distorts along intensity edges
	and therefore preserves edges (or, depending on its mode,
	even anhances edges)
	Proposed by Zhang & Allebach, IEEE Trans Imag Proc 2007; 17: 664

	Works on 2D on a slice-by-slice basis

*/


/* abf_convolve is the core of the filter. It operates on the entire image
	on a 2D slice-by-slice basis. ngh is the neighborhood (support) size.
	sigma_d and sigma_r are the standard deviations for the domain and range
	filters, respectively, mode is the filter mode as follows:
			0: non-adaptive bilateral filter
			1: adaptive, dragging values towards neighborhood mean (smoothing)
			2: adaptive, dragging values away from neighborhood mean (edge-restoring)
			3: adaptive, over-sharpening
*/



void abf_convolve (image_type *img, int ngh, float sigma_d, float sigma_r, int mode)
{
image_type slice;			/* Intermediate storage for filtered slice */
float *p;
float *dkern, *rkern;		/* domain and range kernels */
int x,y,z,x1,y1,m;
float imax, imin, imean;	/* Used for the range filter */
float min, max;				/* Used for normalization */
float normfac;				/* Also for normalization */
float delta, chi;			/* Used for the range filter */
float weight;				/* Normalization factor */
float buf, buf2;
unsigned char normalize;	/* Whether to normalize global values for the range filter */


	if (allocate_image (&slice, B_FLOAT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
	p = slice.data;

	/* Allocate memory for the kernels */
	
	m = 2*ngh+1;		/* Neighborhood total size */
	chi = 0.0;			/* silence a compiler warning */
	sigma_d = 2*SQR(sigma_d);		/* More efficient in kern computation */
	sigma_r = 2*SQR(sigma_r);
	normalize = 0;
	if (mode & 0x80) normalize=1;
	mode &= 0x7F;
	if (img->imgtype==B_FLOAT)
	{
		min = img->minval.fmin; max = img->maxval.fmax;
	}
	else
	{
		min = img->minval.imin; max = img->maxval.imax;
	}
	if (min==max)
		normfac=256;
	else
		normfac = 256.0/(max-min);


	dkern = calloc (SQR(m), sizeof (float));
	rkern = calloc (SQR(m), sizeof (float));
	if (!dkern || !rkern) fatal_memerr();

	/* The Gaussian domain kernel can be precomputed */

	for (y=-ngh; y<=ngh; y++)
		for (x=-ngh; x<=ngh; x++)
		{
			dkern[ (x+ngh) + (y+ngh)*m] = exp (- (SQR(x)+SQR(y))/sigma_d);
		}

	dp (2, "Here is 1/4 of your domain filter kernel for 2*sigmasqr=%f:\n",sigma_d);
	for (y=0; y<=ngh; y++)
	{
		for (x=0; x<=ngh; x++)
			dp (2, "%5.2f  ",dkern[ (x+ngh) + (y+ngh)*m]);
		dp (2, "\n");
	}
	dp (2, "---------------------------------\n");


	for (z=0; z<img->zmax; z++)				/* Slice-by-slice */
	{

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{

				/* First step, compute the adaptive range filter kernel */
	
				imax = imin = readbuf_flt (img,x,y,z); imean = 0;
				for (y1=y-ngh; y1<=y+ngh; y1++)
					for (x1=x-ngh; x1<=x+ngh; x1++)			/* Determine mean, min, max of neighborhood */
					{
						buf = readbuf_flt (img,x1,y1,z);
						imean += buf;
						if (imax<buf) imax=buf;
						if (imin>buf) imin=buf;
					}
				imean /= SQR(m);
				buf = readbuf_flt (img,x,y,z);
				delta = buf - imean;
				switch (mode)				/* Determine range offset chi */
				{
					case 0:					/* Non-adaptive */
						chi = 0;
						break;
					case 1:					/* Shift histo towards mean, smoothing */
						chi = -delta;
						break;
					case 2:					/* Shift histo away from mean, edge-restoring */
						chi = delta;
						break;
					case 3:					/* Separate the bimodal histo, over-sharpening */
						if (delta > 0)
							chi = imax - buf;
						else if (delta < 0)
							chi = imin - buf;
						else
							chi = 0;
						break;
				}

				for (y1=-ngh; y1<=ngh; y1++)			/* Compute the range kernel */
					for (x1=-ngh; x1<=ngh; x1++)
					{
						buf2 = readbuf_flt (img,x+x1,y+y1,z);
						if  (normalize)
						{
							buf2 = (buf2-min)*normfac - (buf-min)*normfac - chi*normfac;	/* Chi is a difference anyway */
						}
						else 
							buf2 -= (buf+chi);

						rkern[ (x1+ngh) + (y1+ngh)*m] = exp (- SQR(buf2)/sigma_r);
					}

				if ((x==84) && (y==86))
				{
					dp (2, "Range values: Imin=%f, Imean=%f, Imax=%f, central pixel =%f, delta=%f\n",imin,imean,imax,buf,delta);
					dp (2, "Here is your complete filter kernel at %d,%d, chi=%f, 2*sigmar^2=%f:\n",x,y,chi,sigma_r);
					for (y1=-ngh; y1<=ngh; y1++)
					{
						for (x1=-ngh; x1<=ngh; x1++)
							dp (2, "%5.3f  ",rkern[ (x1+ngh) + (y1+ngh)*m]*dkern[ (x1+ngh) + (y1+ngh)*m]);
						dp (2, "\n");
					}
					dp (2, "---------------------------------\n");
				}

				/* Second step, apply the kernel to compute target pixel at (x,y) */

				weight = 0.0; buf = 0.0;
				for (y1=-ngh; y1<=ngh; y1++)
					for (x1=-ngh; x1<=ngh; x1++)
					{
						buf += readbuf_flt (img, x+x1,y+y1,z) * dkern[ (x1+ngh) + (y1+ngh)*m] * rkern[ (x1+ngh) + (y1+ngh)*m];
						weight += dkern[ (x1+ngh) + (y1+ngh)*m] * rkern[ (x1+ngh) + (y1+ngh)*m];
					}

				if (weight ==0.0) buf = 0; else buf /= weight;
				p[x + img->xmax*y] = buf;
			}
			
		/* Third step, store filtered slice back in source image */

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = p[x + img->xmax*y];
				switch (img->imgtype)
				{
					case B_CHAR:
					{
						unsigned char* q = img->data;
						int c = (int) (0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						q[x + img->xmax*(y + img->ymax*z)]=c;
						break;
					}
					case B_SHORT:
					{
						short* q = img->data;
						int s = (int) (0.5+buf);
						q[x + img->xmax*(y + img->ymax*z)]=s;
						break;
					}
					case B_FLOAT:
					{
						float* q = img->data;
						q[x + img->xmax*(y + img->ymax*z)]=buf;
						break;
					}
					case B_RGB:
					{
						rgbtriplet* q = img->data;
						int c = (int) (0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						q[x + img->xmax*(y + img->ymax*z)]=(rgbtriplet){c,c,c};
						break;
					}
				}
			}

	}

	freebuf (&slice);
	make_minmax (img);
	img->uid = next_uid++;

}


void abf_main (int ngh, float sigmad, float sigmar, int mode)
{

	dp (1, "Starting adaptive bilateral filter: Sigma D=%4.1f, Sigma R = %4.1f, Mode=%d, Neighborhood=%d\n",sigmad,sigmar,mode, ngh);
	abf_convolve (&mainimg, ngh, sigmad, sigmar, mode);
		
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
}



/*************************************************************

	Adaptive Filter GUI
	
*************************************************************/

GtkAdjustment* afilter_ngh_adj;
GtkAdjustment *afilter_thresh_adj, *afilter_iter_adj;
GtkAdjustment *afilter_sigmar_adj, *afilter_sigmad_adj;
GtkAdjustment *afilter_gamma_adj;
GtkWidget *afilter_ngh, *afilter_thresh, *afilter_iter;
GtkWidget *afilter_sigmar, *afilter_sigmad, *afilter_gamma;
GtkWidget *afilter_label3, *afilter_label4, *afilter_label5, *afilter_label6, *afilter_label7;
GtkWidget *afilter_label8, *afilter_label9, *afilter_label10;
GtkWidget *abf_optionmenu1;
GtkWidget *abf_normalize_button;

int afilter_function, abf_function, abf_normalize;


void on_afilter_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_afilter_ok_clicked  (GtkButton *button, gpointer user_data)
{
int nghsize, iter;
int norm;
float thresh,sigmad,sigmar, gamma;

	undo_nexus (&mainimg);

	nghsize = afilter_ngh_adj->value;
	iter = afilter_iter_adj->value;
	thresh = afilter_thresh_adj->value;
	gamma = afilter_gamma_adj->value;

	if (afilter_function==0)
		local_adaptive_lowpass_main(nghsize, gamma);
	else if (afilter_function==1)
		tristate_median_main(thresh);
	else if (afilter_function==2)
		anisotropic_diffusion_main (thresh, iter);
	else if (afilter_function==3)
	{
		norm = gtk_toggle_button_get_active ((GtkToggleButton*) abf_normalize_button);
		if (norm) norm = 0x80;
		dp (2, "ABF: Normalize is %d\n", norm);
		sigmar = afilter_sigmar_adj->value;
		sigmad = afilter_sigmad_adj->value;
		abf_main (nghsize, sigmad, sigmar, abf_function | norm);
	}
	
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}

void afilter_set_menu (GtkWidget *widget, gpointer user_data)
{

	afilter_function = *(int*) user_data;
	switch (afilter_function)
	{
		case 0:		/* Local adaptive minimum mean-squared error (MMSE) filter */
		{
			gtk_widget_show (afilter_label10);
			gtk_widget_show (afilter_gamma);
			gtk_widget_hide (afilter_label3);
			gtk_widget_show (afilter_ngh);
			gtk_widget_hide (afilter_label4);
			gtk_widget_hide (afilter_thresh);
			gtk_widget_hide (afilter_label5);
			gtk_widget_hide (afilter_iter);
			gtk_widget_hide (afilter_label6);
			gtk_widget_hide (afilter_sigmar);
			gtk_widget_hide (afilter_label7);
			gtk_widget_hide (afilter_sigmad);
			gtk_widget_hide (afilter_label8);
			gtk_widget_hide (abf_optionmenu1);
			gtk_widget_hide (afilter_label9);
			gtk_widget_hide (abf_normalize_button);
			break;
		}
		case 1:		/* Tristate median filter */
		{
			gtk_label_set_text (GTK_LABEL(afilter_label4), "Threshold");
			gtk_widget_hide (afilter_ngh);
			gtk_widget_hide (afilter_label10);
			gtk_widget_hide (afilter_gamma);
			gtk_widget_show (afilter_label3);
			gtk_widget_show (afilter_label4);
			gtk_widget_show (afilter_thresh);
			gtk_widget_hide (afilter_label5);
			gtk_widget_hide (afilter_iter);
			gtk_widget_hide (afilter_label6);
			gtk_widget_hide (afilter_sigmar);
			gtk_widget_hide (afilter_label7);
			gtk_widget_hide (afilter_sigmad);
			gtk_widget_hide (afilter_label8);
			gtk_widget_hide (abf_optionmenu1);
			gtk_widget_hide (afilter_label9);
			gtk_widget_hide (abf_normalize_button);
			break;
		}
		case 2:		/* Anisotropic diffusion filter */
		{
			gtk_label_set_text (GTK_LABEL(afilter_label4), "Diffusion Strength");
			gtk_widget_hide (afilter_ngh);
			gtk_widget_hide (afilter_label10);
			gtk_widget_hide (afilter_gamma);
			gtk_widget_show (afilter_label3);
			gtk_widget_show (afilter_label4);
			gtk_widget_show (afilter_thresh);
			gtk_widget_show (afilter_label5);
			gtk_widget_show (afilter_iter);
			gtk_widget_hide (afilter_label6);
			gtk_widget_hide (afilter_sigmar);
			gtk_widget_hide (afilter_label7);
			gtk_widget_hide (afilter_sigmad);
			gtk_widget_hide (afilter_label8);
			gtk_widget_hide (abf_optionmenu1);
			gtk_widget_hide (afilter_label9);
			gtk_widget_hide (abf_normalize_button);
			break;
		}
		case 3:		/* Adaptive Bilateral Filter */
		{
			gtk_widget_hide (afilter_label3);
			gtk_widget_show (afilter_ngh);
			gtk_widget_hide (afilter_label10);
			gtk_widget_hide (afilter_gamma);
			gtk_widget_hide (afilter_label4);
			gtk_widget_hide (afilter_thresh);
			gtk_widget_hide (afilter_label5);
			gtk_widget_hide (afilter_iter);
			gtk_widget_show (afilter_label6);
			gtk_widget_show (afilter_sigmar);
			gtk_widget_show (afilter_label7);
			gtk_widget_show (afilter_sigmad);
			gtk_widget_show (afilter_label8);
			gtk_widget_show (abf_optionmenu1);
			gtk_widget_show (afilter_label9);
			gtk_widget_show (abf_normalize_button);
			break;
		}

	}

}

void abf_set_menu (GtkWidget *widget, gpointer user_data)
{
	abf_function = *(int*) user_data;

}




void create_afilter_wnd (image_type* img)
{
  GtkWidget *afilter_wnd;
  GtkWidget *vbox1;
  GtkWidget *table1;
  GtkWidget *afilter_label1, *afilter_label2;
  GtkWidget *afilter_optionmenu1;
  GtkWidget *afilter_optionmenu1_menu;
  GtkWidget *abf_optionmenu1_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *hbox1, *hsep1;
  GtkWidget *afilter_okbutton;
  GtkWidget *afilter_cancelbutton;
  GtkTooltips *tt_ngh, *tt_thresh, *tt_iter, *tt_sigmad, *tt_sigmar, *tt_agamma;


  float min, max;
  int y;

	afilter_function=0;
	if (img->imgtype==B_FLOAT)
	{
		min = img->minval.fmin; max = img->maxval.fmax;
	}
	else
	{
		min = img->minval.imin; max = img->maxval.imax;
	}

	afilter_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (afilter_wnd), "afilter_wnd", afilter_wnd);
	gtk_window_set_title (GTK_WINDOW (afilter_wnd), "Adaptive Spatial-Domain Filters");

	vbox1 = vbox_in_window (afilter_wnd,"vbox1"); 
	table1 = MakeTable (afilter_wnd, vbox1, "table1", 2,8);

	y=0;
	afilter_label1 = MakeLabel (afilter_wnd, table1, "Filter type", "label1", 0,y);
	afilter_optionmenu1 = Make_Optionmenu (afilter_wnd, table1,"afilter_menu",1,y);
	gtk_container_set_border_width (GTK_CONTAINER (afilter_optionmenu1), 1);
	afilter_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (afilter_optionmenu1_menu, GTK_SIGNAL_FUNC(afilter_set_menu), "Local adaptive lowpass (AMMSE)", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (afilter_optionmenu1_menu, GTK_SIGNAL_FUNC(afilter_set_menu), "Tristate Median Filter", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (afilter_optionmenu1_menu, GTK_SIGNAL_FUNC(afilter_set_menu), "Anisotropic Diffusion", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (afilter_optionmenu1_menu, GTK_SIGNAL_FUNC(afilter_set_menu), "Adaptive Bilateral Filter", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (afilter_optionmenu1), afilter_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (afilter_optionmenu1), 0);

	y++;
	afilter_label2 = MakeLabel (afilter_wnd, table1, "Neighborhood size", "afilter_label2", 0,y);
	afilter_ngh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 1, 13, 1, 1, 0));
	afilter_ngh = MakeSpinbutton (afilter_ngh_adj, afilter_wnd, table1,"afilter_ngh", 1,y);
	gtk_widget_show (afilter_ngh);
	tt_ngh = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_ngh, afilter_ngh, "Neighborhood radius r.\nNeighborhood is (2r+1)x(2r+1)", NULL);
	afilter_label3 = MakeLabel (afilter_wnd, table1, "(always 3x3)", "afilter_label3", 1,y);
	gtk_widget_hide (afilter_label3);

	y++;
	afilter_label10 = MakeLabel (afilter_wnd, table1, "Gamma", "afilter_label10", 0,y);
	afilter_gamma_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 1, 3, 0.1, 1, 0));
	afilter_gamma = MakeSpinbutton (afilter_gamma_adj, afilter_wnd, table1,"afilter_gam", 1,y);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON(afilter_gamma), 1);
	tt_agamma = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_agamma, afilter_gamma, "Gamma correction (strengthens filter)", NULL);

	y++;
	afilter_label4 = MakeLabel (afilter_wnd, table1, "Threshold", "afilter_label4", 0,y);
	afilter_thresh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, min, max, 1, 1, 0));
	afilter_thresh = MakeSpinbutton (afilter_thresh_adj, afilter_wnd, table1,"afilter_thr", 1,y);
	gtk_widget_hide (afilter_label4);
	gtk_widget_hide (afilter_thresh);
	tt_thresh = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_thresh, afilter_thresh, "Diffusion strength k\nOr median threshold", NULL);

	y++;
	afilter_label5 = MakeLabel (afilter_wnd, table1, "Iterations", "afilter_label5", 0,y);
	afilter_iter_adj = GTK_ADJUSTMENT(gtk_adjustment_new (20, 1, 300, 1, 1, 0));
	afilter_iter = MakeSpinbutton (afilter_iter_adj, afilter_wnd, table1,"afilter_iter", 1,y);
	gtk_widget_hide (afilter_label5);
	gtk_widget_hide (afilter_iter);
	tt_iter = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_iter, afilter_iter, "Number of iterations", NULL);

	y++;
	afilter_label6 = MakeLabel (afilter_wnd, table1, "Domain Filter Sigma", "afilter_label6", 0,y);
	afilter_sigmad_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0.1, 20, 0.1, 1, 0));
	afilter_sigmad = MakeHScale (afilter_sigmad_adj, afilter_wnd, table1, "afilter_sigmad", 1,y);
	gtk_scale_set_digits (GTK_SCALE (afilter_sigmad), 1);
	gtk_widget_hide (afilter_label6);
	gtk_widget_hide (afilter_sigmad);
	tt_sigmad = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sigmad, afilter_sigmad, "Domain filter sigma\n(lager = stronger)", NULL);
	
	y++;
	afilter_label7 = MakeLabel (afilter_wnd, table1, "Range Filter Sigma", "afilter_label7", 0,y);
	afilter_sigmar_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0.1, 20, 0.1, 1, 0));
	afilter_sigmar = MakeHScale (afilter_sigmar_adj, afilter_wnd, table1, "afilter_sigmar", 1,y);
	gtk_scale_set_digits (GTK_SCALE (afilter_sigmar), 1);
	gtk_widget_hide (afilter_label7);
	gtk_widget_hide (afilter_sigmar);
	tt_sigmar = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sigmar, afilter_sigmar, "Range filter sigma\n(lager = blurs stronger discontinuities)", NULL);

	y++;
	afilter_label8 = MakeLabel (afilter_wnd, table1, "Range Filter Mode", "afilter_label8", 0,y);
	abf_optionmenu1 = Make_Optionmenu (afilter_wnd, table1,"abf_menu",1,y);
	gtk_container_set_border_width (GTK_CONTAINER (afilter_optionmenu1), 1);
	abf_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (abf_optionmenu1_menu, GTK_SIGNAL_FUNC(abf_set_menu), "Mode 0: Nonadaptive bilateral", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (abf_optionmenu1_menu, GTK_SIGNAL_FUNC(abf_set_menu), "Mode 1: Smoothing", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (abf_optionmenu1_menu, GTK_SIGNAL_FUNC(abf_set_menu), "Mode 2: Edge enhancing", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (abf_optionmenu1_menu, GTK_SIGNAL_FUNC(abf_set_menu), "Mode 3: Over-sharpening", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (abf_optionmenu1), abf_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (abf_optionmenu1), 0);
	gtk_widget_hide (afilter_label8);
	gtk_widget_hide (abf_optionmenu1);
	abf_function = 0;

	y++;
	afilter_label9 = MakeLabel (afilter_wnd, table1, "Range Filter:", "afilter_label9", 0,y);
	abf_normalize_button = Checkbutton_in_table (afilter_wnd, table1, "Normalize value range", "abf_normalize_button", 1 , y, NULL);
	gtk_widget_hide (afilter_label9);
	gtk_widget_hide (abf_normalize_button);
	abf_normalize = 0;
 

	hsep1 = MakeHsep (afilter_wnd, vbox1, "hsep1", -1,-1);
	hbox1 = hbox_in_window (afilter_wnd, vbox1, "hbox1");

	afilter_okbutton = Button_in_table (afilter_wnd, hbox1, "OK", "afilter_okbutton",-1,-1,NULL);
	afilter_cancelbutton = Button_in_table (afilter_wnd, hbox1, "Cancel", "afilter_cancelbutton",-1,-1,NULL);

	gtk_signal_connect (GTK_OBJECT (afilter_okbutton), "clicked", GTK_SIGNAL_FUNC (on_afilter_ok_clicked), afilter_wnd);
	gtk_signal_connect (GTK_OBJECT (afilter_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_afilter_cancel_clicked), afilter_wnd);
	gtk_signal_connect (GTK_OBJECT (afilter_wnd), "destroy", GTK_SIGNAL_FUNC (on_afilter_cancel_clicked), afilter_wnd);

	gtk_widget_show (afilter_wnd); 

}



/* and the menu function for the convolution filters ... */

void afilter_filter_main ()
{

	create_afilter_wnd(&mainimg);

}


