

/*
 *	Morph Op: Morphological operators
 *	- Erode and dilate, open, close
 * 	- Skeletonize
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "application.h"
#include "backgnd.h"
#include "defs.h"
#include "dlgbox.h"
#include "filter.h"
#include "findclust.h"
#include "functions.h"
#include "functions2.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "morphop.h"
#include "roi.h"
#include "secondwnd.h"
#include "special1.h"
#include "tiffstuff.h"
#include "viz.h"





/******************************************************************************************


  ###    #   #   #####   #       #####   #####    ###    #   #    ###    #####   #####  
 #   #   #  #    #       #       #         #     #   #   #   #     #         #   #      
 #       # #     #       #       #         #     #   #   ##  #     #        #    #      
  ###    ##      ####    #       ####      #     #   #   # # #     #       #     ####   
     #   # #     #       #       #         #     #   #   #  ##     #      #      #      
 #   #   #  #    #       #       #         #     #   #   #   #     #     #       #      
  ###    #   #   #####   #####   #####     #      ###    #   #    ###    #####   #####  

/******************************************************************************************/






/**************************************************

	Skeletonize: Zhang & Suen Algorithm
	(Bankmann pg. 221-223)
	The Zhang-Suen algorithm yields an 8-connected
	skeleton.
	Fate table "lifted" from NIH Image, Filter.p
	Another fate table generated from data in 
	Ritter GX, Wilson JN, Computer Vision Algorithms
	in Image Algebra, CRC 1996
	
***************************************************/


/* These tables define the skeletonization neighborhood.
   The displacements have the following directions (one more than needed) 

	0  1  2  3  4  5  6  7  8
	N  NE E  SE S  SW W  NW N    (clockwise starting at 12 o'clock)
	
	and the binval weights each bit so that we can merge it into
	a uchar offset into a fate table. We start with 1 at NW and run
	clockwise (Note that Zhang & Suen start with the North pixel)
*/

int skel_xdisp[9] = { 0, 1, 1, 1, 0,-1,-1,-1, 0};	/* x-offsets for direction */
int skel_ydisp[9] = {-1,-1, 0, 1, 1, 1, 0,-1,-1};
int skel_binval[9]= { 2, 4, 8,16,32,64,128,1, 2};	/* Binary values of directions for fate table index, NIH */
int skel_Binval[9]= { 1, 2, 4, 8,16,32,64,128,1};	/* The same, Ritter & Wilson and Zhang & Suen */



/*	We use the same scheme as for the EDM (see special.c)
	Thus, we call this part indirectly from the threshold window.

	skel_iterate expects a binary image (B_CHAR) and performs
	one iteration (two steps) of the Zhang-Suen thinning.
	the skeletonization works in 2D only, thus the image
	passed to skel_iterate is one slice only.
	the caller of skel_iterate() is responsible for providing
	a 2D uchar field to store the "marked for deletion" flags.
	
	Edge handling: Whatever is off the image is zero.
	
	Some general observations. Most thinning algorithms (including
	Zhang & Suen) assume 8-conectivity, but fail insofar as they leave
	over plenty of 4-connected pixels that should have been deleted.
	This includes many commercial / professional programs. The NIH Image
	fate table is one of the best approximations available and removes
	more 4-connected pixels than Zhang & Suen as described in Bankmann or
	Ritter & Wilson. I added an additional tidyup step that removes several
	typical 4-connected pixels. 
	
*/

void skel_thresh (image_type* src, image_type* dest, int slice, float thresh)
{
rgbtriplet rgb;
rgbtriplet* sp;
unsigned char* p;
int x,y;

	p = dest->data; dest->minval.imin=0; dest->maxval.imax=1;
	
	if (src->imgtype==B_RGB)
	{
		sp = src->data;
		for (x=0; x<src->xmax; x++)
			for (y=0; y<src->ymax; y++)
			{
				rgb = sp[x + src->xmax*y];
				p[x + dest->xmax*y] = 0;
				if ((rgb.red >= thresh) || (rgb.green >= thresh) || (rgb.blue >= thresh))
					p[x + dest->xmax*y] = 1;
				else
					p[x + dest->xmax*y] = 0;
			}
	}
	else
	{
		for (x=0; x<src->xmax; x++)
			for (y=0; y<src->ymax; y++)
				p[x + dest->xmax*y] = (readbuf_flt(src,x,y,slice)>=thresh ? 1 : 0 );
	}
}


/* Return whether a pixel at (x,y) is nonzero */

char skel_getval (image_type* img, int x, int y)
{
char val;

	val=0;
	if ((x<0) || (y<0) || (x>=img->xmax) || (y>=img->ymax)) return 0;
	
	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = img->data;
			val = (p[x + img->xmax*y])>0;
			break;
		}
		case B_SHORT:
		{
			short* p = img->data;
			val = (p[x + img->xmax*y])!=0;
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data;
			val = (p[x + img->xmax*y])!=0.0;
			break;
		}
		case B_RGB:
		{
			rgbtriplet *rgb = img->data;
			val = ((rgb[x + img->xmax*y].red!=0) || (rgb[x + img->xmax*y].green!=0) || (rgb[x + img->xmax*y].blue!=0));
			break;
		}
	}
	return val;
}


/* Determine the number of 8-connected nonzero neighbors of a pixel (x0,y0).
   Works on B_CHAR and B_SHORT images ONLY!!  8-connectivity ONLY!! */

int skel_numneigh (image_type* img, int x0, int y0)
{
int c,x,y;

	c =0; 

	if (img->imgtype==B_CHAR)
	{
		unsigned char* pu = img->data;
		for (y = MAX(0,y0-1); y <= MIN(img->ymax-1,y0+1); y++)
			for (x = MAX(0,x0-1); x <= MIN(img->xmax-1,x0+1); x++)
			{
				if ((x!=x0) || (y!=y0)) 
				{
					if ( pu[x + img->xmax*y] ) c++;
				}
			}
	}
	else if (img->imgtype==B_SHORT)
	{	
		short* ps = img->data;
		for (y = MAX(0,y0-1); y <= MIN(img->ymax-1,y0+1); y++)
			for (x = MAX(0,x0-1); x <= MIN(img->xmax-1,x0+1); x++)
			{
				if ((x!=x0) || (y!=y0)) 
				{
					if ( ps[x + img->xmax*y] ) c++;
				}
			}
	}
	else
		ShowMessage ("Illegal operation");
	
	return c;
}
	

/* Check for a specific value in the neighboorhood of (x,y) */

int skel_checkforval (image_type* img, int x0, int y0, int val)
{
int x,y,c;
unsigned char* pu;
short* ps;

	c=0; pu = img->data; ps = img->data;
	for (y = MAX(0,y0-1); y <= MIN(img->ymax-1,y0+1); y++)
		for (x = MAX(0,x0-1); x <= MIN(img->xmax-1,x0+1); x++)
		{
			if (img->imgtype==B_CHAR) c=pu[x + img->xmax*y];
			else if (img->imgtype==B_SHORT) c=ps[x + img->xmax*y];
			if (c==val) return 1;
		}
	
	return 0;
}


/* This algorithm converts the Zhang-Suen conditions into a fate table. 
   It reproduces *exactly* the table on page 153 of Ritter GX, Wilson JN,
   Computer Vision Algorithms in Image Algebra, CRC 1996 */

void make_fatetbl ()
{
int i,j,k,n;
unsigned char m[9];	/* a 3x3 matrix */
unsigned char offs[9] = {1,2,5,8,7,6,3,0,1};
unsigned char pass1, pass2;
unsigned char res[256];

	for (i=0; i<256; i++)
	{
		/* Fill up the matrix with the neighborhood pattern */
		
		for (j=0; j<9; j++) m[j]=0;
		k = i;
		for (j=0; j<8; j++)
		{
			if (k & 0x01) m[offs[j]]=1;
			k = k>>1;
		}
		
		/* Check the Zhang-Suen conditions */
		
		pass1=1; pass2=1;
		n=0; for (j=0; j<8; j++) if (m[offs[j]]) n++;	/* Number of neighbors */
		if ((n<2) || (n>6))
		{
			pass1=0; pass2=0;		/* Endpoints not preserved */
		}
		n=0; 
		for (j=0; j<8; j++) 
			if ((m[offs[j]]==0) && (m[offs[j+1]]==1)) n++; 	/* 0 to 1 transitions */
		if (n!=1)					/* Point between endpoints */
		{
			pass1=0; pass2=0;
		}
		
		if ( m[offs[0]]*m[offs[2]]*m[offs[4]] !=0) pass1=0;
		if ( m[offs[2]]*m[offs[4]]*m[offs[6]] !=0) pass1=0;
		if ( m[offs[0]]*m[offs[2]]*m[offs[6]] !=0) pass2=0;
		if ( m[offs[0]]*m[offs[4]]*m[offs[6]] !=0) pass2=0;
		
		dp (0, "Pattern %3d (%02x) Pass1 %1d, pass2 %1d\n",i,i,pass1,pass2);
		res[i] = pass1 + 2*pass2;
		
	}

	for (i=0; i<256; i++)
	{
		if ( !(i & 0x1f)) dp (0,"\n");
		dp (0,"%1d,",res[i]);
	}
	dp (0,"\n");

}


/* As above, but for Haidekker, Advanced Biomedical Image Analysis, Chapter 9.
	This algorithm is an *exact* representation of the one presented by
	Zhang & Suen, "A fast parallel algorithm for thinning digital patterns",
	Communications of the ACM, 1984; 27(3): 236-239.
	Note - if you set compat=1, the table is shifted to start with the
	NW corner rather than the North neighbor, in compatibility with 
	NIH Image.
	Note that this algorithm does *not* provide a working table -- the
	resulting fate table seems to be too aggressive */

void make_fatetbl2 ()
{
int i,j,k,n,compat;
unsigned char m[9];
unsigned char pass1, pass2;
unsigned char res[256];

	compat=1;

	for (i=0; i<256; i++)
	{
		/* Fill up the matrix with the neighborhood pattern.
			Note offset to Zhang & Suen. In their case, North is 2.
			In our case, North is 0 */
		
		for (j=0; j<9; j++) m[j]=0;
		if (compat==0)
		{
			k = i;
			for (j=0; j<8; j++)
			{
				if (k & 0x01) m[j]=1;
				k = k>>1;				/* North goes into LSB, then clockwise */
			}
		}
		else
		{
			for (j=0; j<8; j++)
				if (i & skel_binval[j]) m[j]=1;
		}
		m[8]=m[0];					/* Cyclic completion to get transitions from 7 to 0 */

		/* Check the Zhang-Suen conditions */
		
		pass1=1; pass2=1;
		n=0; for (j=0; j<8; j++) if (m[j]) n++;		/* Number of neighbors */
		if ((n<2) || (n>6))			/* Zhang & Suen condition (a) */
		{
			pass1=0; pass2=0;		/* Endpoints not preserved */
		}
		
		n=0; 
		for (j=0; j<8; j++) 
			if ((m[j]==0) && (m[j+1]!=0)) n++; 	/* 0 to 1 transitions */
		if (n!=1)					/* Zhang & Suen condition (b) */
		{
			pass1=0; pass2=0;		/* Point between endpoints */
		}
		
		if ( m[0] && m[2] && m[4]) pass1=0;		/* Zhang & Suen condition (c) */
		if ( m[2] && m[4] && m[6]) pass1=0;		/* Zhang & Suen condition (d) */
		if ( m[0] && m[2] && m[6]) pass2=0;		/* Zhang & Suen condition (c') */
		if ( m[0] && m[4] && m[6]) pass2=0;		/* Zhang & Suen condition (d') */
		
		dp (0, "Pattern %3d (%02x) Pass1 %1d, pass2 %1d\n",i,i,pass1,pass2);
		res[i] = pass1 + 2*pass2;
		
	}

	for (i=0; i<256; i++)
	{
		if ( !(i & 0x1f)) dp (0,"\n");
		dp (0,"%1d,",res[i]);
	}
	dp (0,"\n");

}


void fatetbl_shuffle ()
{

char fatetbl[256] = {	0,0,0,1,0,0,1,3,0,0,3,1,1,0,1,3,0,0,0,0,0,0,0,0,2,0,2,0,3,0,3,3, /* 00-31 */
						0,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,2,0,0,0,3,0,2,2, /* 32-63 */
						0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 64-95 */
						2,0,0,0,0,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,0,0,0,0,3,0,0,0,3,0,2,0, /* 96-127 */
						0,1,3,1,0,0,1,3,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1, /* 128-159 */
						3,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 160-191 */
						2,3,1,3,0,0,1,3,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 192-223 */
						2,3,0,1,0,0,0,1,0,0,0,0,0,0,0,0,3,3,0,1,0,0,0,0,2,2,0,0,2,0,0,0  /* 224-255 */
					};

int i,j;


	for (i=0; i<256; i++)
	{
		j = (i<<1) & 0xFE;
		if (i & 0x80) j=j | 1;
		if ( !(i & 0x1f)) dp (0,"\n");
		dp (0,"%1d,",fatetbl[j]);
	}
	dp (0,"\n");


}



/* The "core" of the skeletonizer. We analyze the neighborhood patterns
   and delete pixels according to a fate table (Bankmann, Russ, NIH Image).
   This is more efficient than the standard Zhang-Suen method where
   plain comparisons are made. Also, the algorithm as described in
   Bankmann does not work correctly. Nor does the one by Zhang & Suen, btw. */

long skel_iterate (image_type* img, unsigned char* p1)
{
int x,y,i,j,pass;
unsigned char* p;
long changes;
unsigned char fateidx;

/* Fate table. 0 means don't remove; 1: Remove in pass 1; 2: remove in pass 2; 3: Remove either pass.
   This is the NIH Image version. Generate the Ritter & Wilson version with make_fatetbl */

char fatetbl[256] = {	0,0,0,1,0,0,1,3,0,0,3,1,1,0,1,3,0,0,0,0,0,0,0,0,2,0,2,0,3,0,3,3, /* 00-31 */
						0,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,2,0,0,0,3,0,2,2, /* 32-63 */
						0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 64-95 */
						2,0,0,0,0,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,0,0,0,0,3,0,0,0,3,0,2,0, /* 96-127 */
						0,1,3,1,0,0,1,3,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1, /* 128-159 */
						3,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 160-191 */
						2,3,1,3,0,0,1,3,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 192-223 */
						2,3,0,1,0,0,0,1,0,0,0,0,0,0,0,0,3,3,0,1,0,0,0,0,2,2,0,0,2,0,0,0  /* 224-255 */
					};

char zsfatetbl[256] = {	/* The above, but reshuffled to match the Zhang-Suen neighborhood definition */
						0,0,0,1,0,3,1,1,0,0,0,0,2,2,3,3,0,0,0,0,3,0,0,0,2,0,0,0,2,0,3,2,
						0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,2,0,2,0,3,0,0,0,3,0,3,2,
						0,3,0,1,0,0,0,0,0,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,
						2,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,3,0,0,0,2,0,2,0,
						0,1,0,3,0,1,0,3,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,
						0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
						1,1,0,3,0,0,0,1,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
						3,3,0,3,0,0,0,1,0,0,0,0,0,0,0,0,3,1,0,1,0,0,0,0,3,1,0,0,2,0,0,0	
					};

					
					
	p = img->data;

	changes=0;
	for (pass=0; pass<=1; pass++)
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
		
				if (p[x + img->xmax*y])	/* can only delete a point that is set */
				{
					
					dp (3, "Pass %d, point (%d,%d): ",pass,x,y);
					
					/* Compute index into fate table from neighbor pattern */
					fateidx=0;
					for (i=0; i<8; i++)
					{
						j = skel_getval (img,x+skel_xdisp[i],y+skel_ydisp[i]);
						if (j) fateidx |= skel_binval[i];
					}
					dp (3, " Fate table entry %d values %d ",fateidx,fatetbl[fateidx]);
					
					/* Mark for deletion according to fate table */
					if ( (fatetbl[fateidx]==3)
							|| ((fatetbl[fateidx]==2) && (pass==1))
							|| ((fatetbl[fateidx]==1) && (pass==0)))
					{
						p1[x + img->xmax*y] = 1; /* mark for deletion */
						dp (3, " *D* ");
					}
						

					dp (3, "\n");
				}
			}
			
		/* After all pixels have been marked for deletion, actually delete them */
		
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (p1[x + img->xmax*y]) 
				{
					p[x + img->xmax*y]=0; p1[x + img->xmax*y]=0;
					changes++;
					dp (3, "Deleting (%d,%d)\n",x,y);
				}
			}
			
	}
	
	/* tell the caller how many pixels we deleted */

	return changes;
}

/* Tidy up. The fate-table driven algorithm leaves a specific pattern
   where the E, S, W, NW, N neighbors are set. We can't enable this in
   the iterative process, because it tears apart the skeleton, so
   we check for 0xE8 neighborhoods in a separate step.
*/


long skel_tidyup (image_type* img)
{
int x,y,i,j,l,fateidx,offend;
unsigned char* p;
long changes;

/* Offending 4-connected patterns. These come in rotated patterns of four */
unsigned char offending[] = {	0xE8, 0xA3, 0x8E, 0x3A,
								0x8B, 0xE2, 0xE8, 0x2E,
								0x8A, 0x2A, 0xA2, 0xA8,
								0x82, 0x0a, 0x28, 0xa0
							};

	p = img->data;
	changes=0;
	l = sizeof (offending) / sizeof (unsigned char);

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{

			if (p[x + img->xmax*y])	/* can only delete a point that is set */
			{

				/* Compute index into fate table from neighbor pattern */
				fateidx=0;
				for (i=0; i<8; i++)
				{
					j = skel_getval (img,x+skel_xdisp[i],y+skel_ydisp[i]);
					if (j) fateidx |= skel_binval[i];
				}

				/* Delete immediately if this is an offending pattern */
				
				offend=0;
				for (i=0; i<l; i++) 
					if (fateidx==offending[i]) offend=1;
				if (offend)
				{
					p[x + img->xmax*y] = 0; /* erase */
					dp (2, "Skel: Offending pattern %02X deleted at %d,%d\n",fateidx,x,y);
					changes++;
				}
			}
		}
			
	
	/* tell the caller how many pixels we deleted */

	return changes;
}



/* The skeletonizer function. It takes any image type
   and uses the threshold parameters in thresh_return
   to create a binary image. It then iteratively calls
   skel_iterate() until the attractor image is found.
   This attractor image (now B_CHAR) replaces the old one
   passed via img.
*/

void skeletonize (image_type* img, int slice, float threshold)
{
image_type dest;
long cnt,i;
unsigned char* p1;
unsigned char* p2;


	if (allocate_image (&dest, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p1 = (unsigned char*) calloc (img->xmax*img->ymax,sizeof(unsigned char));
	if (!p1) fatal_memerr();

	skel_thresh (img, &dest, slice, threshold);

	i=0;
	do
	{
		cnt = skel_iterate (&dest, p1); i++;
		dp (1, "Skeletonize: Iteration %li, points deleted: %li\n",i,cnt);
	}
	while (cnt>0);

	cnt = skel_tidyup (&dest);
	dp (1, "Skeletonize: Tidyup, points deleted: %li\n",cnt);

	p2 = dest.data;
	for (i=0; i<dest.xmax*dest.ymax; i++) if (p2[i]) p2[i]=255;

	/* Return dest in img */

	handback (img, &dest);
	img->minval.imin=0; img->maxval.imax=255; cur_slice=0;
	free (p1);
}



/* 
	Prune any "fuzz" of length <=n.
	Work on skeletonized image before labeling in skel_statistics.
	Algorithm
		1. Make a copy of the original image
		2. For n iterations, erode endpoints only (in copy)
			(this means that e.g. links, nodes cannote be eroded)
		3. For n iterations, dilate the copy under the constraints
			(a) pixel is an endpoint (0 or 1 neighbor)
			(b) neighbor to be set was also set in original image
		4. Return copy, discard original
*/


void skel_prune (image_type *img, int prunelen)
{
image_type tempimg,tempimg2;
unsigned char *p; unsigned char *p2;
int x,y,x0,y0,round,xm,n,prunecnt;


	if (prunelen <=0) return;

	/* Create a copy of the original skeletonized image (img) to work on */

	if (allocate_image (&tempimg, B_CHAR, img->xmax, img->ymax,1)==-1)
		fatal_memerr();
	if (allocate_image (&tempimg2, B_CHAR, img->xmax, img->ymax,1)==-1)
		fatal_memerr();
	p = tempimg.data;
	p2=tempimg2.data;
	tempimg.minval.imin=0; tempimg.maxval.imax=1;
	tempimg2.minval.imin=0; tempimg2.maxval.imax=1;
	xm=img->xmax;
	prunecnt=0;

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
			if (readbuf_char (img,x,y,0)!=0)
			{
				p[x+xm*y]=1;
			}



	/* Erode endpoints, multiple passes */
	
	for (round=0; round < prunelen; round++)
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				p2[x+xm*y]=0;
				if (p[x+xm*y])
				{
					n = skel_numneigh (&tempimg,x,y);
					if (n==1) p2[x+xm*y]=255;				/* Mark for deletion */
				}
			}
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (p2[x+xm*y]) 
				{
					p[x+xm*y]=0;		/* actually delete */
					prunecnt++;
					dp (2, "Prune: Endpoint at %d, %d deleted\n",x,y);
				}
			}
	}
	/*secondary_display (&tempimg2,0,"Prune");*/

	/* Dilate endpoints when the underlying original point was set */
	for (round=0; round < prunelen; round++)
	{
		for (y=1; y<img->ymax-1; y++)
			for (x=1; x<img->xmax-1; x++)
				p2[x+xm*y]=0;

		for (y=1; y<img->ymax-1; y++)
			for (x=1; x<img->xmax-1; x++)
			{
				n = skel_numneigh (&tempimg,x,y);
				if ((p[x+xm*y]) && (n==1))
				{
					for (y0=y-1; y0<=y+1; y0++)
						for (x0=x-1; x0<=x+1; x0++)
						{
							if ((readbuf_char(img,x0,y0,0)!=0)
								&& (p[x0+xm*y0]==0))
							{
								p2[x0+xm*y0] = 1;		/* Flag for dilation */
								dp (2, "Prune: Endpoint at %d, %d restored\n",x0,y0);
							}
						}
				}
			}
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (p2[x+xm*y]) 
				{
					p[x+xm*y]=1;		/* actually dilate */
					prunecnt--;
				}
			}
	}

	dp (1, "Skeleton: %d pixels pruned.\n",prunecnt);
	skel_tidyup (&tempimg);

	handback (img, &tempimg);
	freebuf (&tempimg2);
	img->minval.imin=0; img->maxval.imax=1; cur_slice=0;

}


/* This function performs statistics on a skeletonized image and determines
   morphological and other parameters.
*/

void skel_statistics (image_type* img, char* report)
{
int x0,y0,x1,y1,x,y,xm,n,i,j,cnt1,cnt2,max;
int numnodes, numends, numbranches, numelems, numlinks, numloops, eulernum, numholes, dust;
short* p; short* p2;
image_type tempimg, tempimg2;
char s[128];
float branchlen, linklen, looplen, tbranchlen, tlinklen;

	/* First of all, we need a temporary image slice, because we
	   need to modify it */
	
	if (allocate_image (&tempimg, B_SHORT, img->xmax, img->ymax,1)==-1)
		fatal_memerr();
	p = tempimg.data;
	xm=img->xmax;
	
	/* Determine the number of elements by cluster counting (threshold==1)
	   At the same time, we fill tempimg with a copy of img, but defined
	   data type. We need to call pass2 only to clean up memory.
	   NOTE - use the more recent binary mode (labelopts==1) in new_pass2
	   to obtain a pure binary image. */
	
	numelems = new_pass1_2d (img, &tempimg, 0, 1, 1);
	new_pass2 (&tempimg, numelems, NULL, 1, 1);
	p = tempimg.data;

	/* Determine the ends and nodes  - we also color-code these dots:
		1: Regular side (unprocessed)
		2: Dust
		3: End
		4: Node
	*/
	
	numbranches=0; numnodes=0; numends=0; dust=0; numlinks=0;
	branchlen=0; linklen=0; numloops=0; looplen=0;
	tbranchlen=0; tlinklen=0;
	
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
			if (p[x+xm*y])
			{
				n = skel_numneigh (&tempimg,x,y);
				if (n==0)
				{
					dust++;			/* Dust particle (isolated dot) */
					p[x+xm*y]=2;
				}
				else if (n==1) 
				{
					numends++;	/* One neighbor -> end */
					p[x+xm*y]=3;	/* Label end point with 3 */
				}
				else			/* more neighbors may indicate a node */
				{
					cnt1=0; cnt2=0;
					for (i=0; i<8;i++)
					{
						j = skel_getval (img,x+skel_xdisp[i],y+skel_ydisp[i]);
						if (j) cnt1++;
						if ((j==0) && (skel_getval (img,x+skel_xdisp[i+1],y+skel_ydisp[i+1]))) cnt2++;
					}
					/* A true node has >=2 neighbors and >=3 transitions */
					
					if ((cnt1>=2) && (cnt2>=3))
					{
						numnodes++;
						p[x+xm*y]=4;		/* Label node pixel as 4 */
					}

				}
			}

	/*	BRANCHES:
		Copy the current image (with labeled ends/nodes) into tmpimg2
		so we can mess with it. 
		At this time, we start "eating up" (zeroing in tmpimg2) the 
		branches starting at the ends until we find a pixel with value 4
		(indicates that we arrived at a node) or without neighbors
		(indicates we arrived at the other end of a lonely link).
		Leave tempimg untouched with the exception that we label
		branches "5".
	*/

	if (clone_image (&tempimg, &tempimg2)==-1) fatal_memerr();
	p2 = tempimg2.data;
	for (y0=0; y0<img->ymax; y0++)
		for (x0=0; x0<img->xmax; x0++)
			if (p2[x0+xm*y0]==3)		/* Found an end pixel - start tracing */
			{
				n = skel_numneigh (&tempimg2,x0,y0); /* This should always return 1 */
				x=x0; y=y0;
				
				branchlen=0;
				
				while ((n==1) && (!skel_checkforval (&tempimg2,x,y,4)))
				{
					p2[x+xm*y]=0;		/* Delete this end pixel */
					if (p[x+xm*y]==1) p[x+xm*y]=5;	/* and label the branch */

					/* where is the next one? */
					i=0;
					while ((i<9) && (!skel_getval(&tempimg2,x+skel_xdisp[i],y+skel_ydisp[i]))) i++;
					x+=skel_xdisp[i]; y+=skel_ydisp[i];	/* go there */
					n = skel_numneigh (&tempimg2,x,y);	/* and check out the new dot */
					branchlen += sqrt (SQR(skel_xdisp[i])+SQR(skel_ydisp[i]));
				}

				/* We drop out of this while loop if either n==0 (another end)
				   or n>1, meaning branch point (node). Label accordingly.
				*/
				
				if (p[x+xm*y]==1) p[x+xm*y]=5;	/* still part of the branch */
				p2[x+xm*y]=0;	/* in the reduced image we need to set a STOP sign as well*/
				
				if (branchlen>0)		/* We don't ignore "trivial" (length 1) branches */
				{
					tbranchlen += branchlen;
					numbranches++;
				}
			}

	if (numbranches>0) tbranchlen/=numbranches;	/* average length of branches */


	/* 	LINKS:
		We will now do the same procedure, iteratively, and starting with
	   "4" nodes. Each link has to have at least one node at its end.
	   Also, each node may have more than 1 originating link.
	   One more trick: In tight loops, the second step may be ambiguous.
	   We need to choose from multiple possibilities so that we minimize
	   curvature. Links are not affected by pruning. */

	for (y0=0; y0<img->ymax; y0++)
		for (x0=0; x0<img->xmax; x0++)
			if (p2[x0+xm*y0]==4)		/* found a node */
			{
				n = skel_numneigh(&tempimg2,x0,y0); /* how many links? */
	
				while (n>0)		/* repeat this for each link */
				{
					x=x0; y=y0; cnt1=0;
					linklen=0;
					p2[x0+xm*y0]=0;	/* Prevent the algorithm from turning back */
					do 
					{
						/* where is the next one? As far away from the initial point
						   as possile */

						x1=0; y1=0; max=0;
						for (i=0; i<9; i++)
						{
							if (skel_getval(&tempimg2,x+skel_xdisp[i],y+skel_ydisp[i]))	/* valid path */
							{
								j = SQR(x+skel_xdisp[i]-x0)+SQR(y+skel_ydisp[i]-y0);
								if (j>max) /* preferred */
								{
									max=j; x1=x+skel_xdisp[i]; y1=y+skel_ydisp[i];
								}
							}
						}
						
						if ((x1!=0) && (y1!=0))	/* Found a point */
						{
							linklen += sqrt(SQR(x1-x)+SQR(y1-y));
							x=x1; y=y1; cnt1++;
							j = p2[x+xm*y];
							if (p2[x+xm*y]==1) p2[x+xm*y]=0;		/* Delete this end pixel */
							if (p[x+xm*y]==1) p[x+xm*y]=6;		/* and label the link */
						}
						else		/* point has been deleted elsewhere */
						{
							j=0;	/* so drop out of the loop */
							if (cnt1==0) { numlinks--; j=-1; }	/* and correct for the link now gone */
							dp (2, "   *** Dangling link after %d steps.\n",cnt1);
						}
						
					}
					while ((j==1) && (!skel_checkforval (&tempimg2,x,y,4)));
					if (j!=-1) dp (2, "   Link %d from %d,%d to %d,%d\n",numlinks,x0,y0,x,y);

					/* We don't ignore "trivial" (length 1) links */
				
					if (linklen>0)
					{
						tlinklen += linklen;
						numlinks++;
					}
					
					p2[x0+xm*y0]=4;	/* Restore the node that we had to delete so that the alg can't walk back */
					n--;	/* Count down for that node */
				}

				p2[x0+xm*y0]=0;	/* Finally delete the node in the reduced image */
			}
	if (numlinks>0) tlinklen /= numlinks;

	/* At this point, there is only one structure remaining, a loop without nodes.
	   We need to grab these separately and, to fulfill the Euler condition,
	   give each of them a pseudo-node. However, we add neither the pseudo-node
	   nor the face itself to the overall stats, becaue it won't change Euler #
	   and # of holes anyway. We count it as loop, though.
	*/

	for (y0=0; y0<img->ymax; y0++)
		for (x0=0; x0<img->xmax; x0++)
			if (p2[x0+xm*y0])		/* found something */
			{
				n = skel_numneigh (&tempimg2,x0,y0);
				x=x0; y=y0;
				
				while (n>0)
				{
					p2[x+xm*y]=0;		/* Delete this end pixel */
					if (p[x+xm*y]==1) p[x+xm*y]=7;	/* and label the loop */
					/* where is the next one? */
					i=0;
					while ((i<9) && (!skel_getval(&tempimg2,x+skel_xdisp[i],y+skel_ydisp[i]))) i++;
					x+=skel_xdisp[i]; y+=skel_ydisp[i];	/* go there */
					looplen += sqrt (SQR(skel_xdisp[i])+SQR(skel_ydisp[i]));
					n = skel_numneigh (&tempimg2,x,y);	/* and check out the new dot */
				}

				if (p[x+xm*y]==1) p[x+xm*y]=7;	/* still part of the loop */
				p2[x+xm*y]=0;	/* in the reduced image we need to set a STOP sign as well*/
				numloops++;
			}
	if (numloops>0) looplen/=numloops;

	/* The number of holes used to be 1-numnodes-numends+numlinks+numbranches, but 
		this computation is wrong, flawed by disconnected branches that have two ends.
		The Euler number used to be numelems-numholes, but this is not right and has been 
		corrected to nodes+loops-branches.
 	*/

	numholes = 1+numlinks-numnodes; 
	eulernum = numnodes+numholes-numlinks;

	freebuf (&tempimg2);

	strcpy (report, "Skeleton statistics:\n");
	sprintf (s, "Number of elements: %d\n",numelems);
	strcat (report,s);
	sprintf (s, "Number of isolated dots: %d\n",dust);
	strcat (report,s);
	sprintf (s, "Number of ends: %d\n",numends);
	strcat (report,s);
	sprintf (s, "Number of branches: %d\n",numbranches);
	strcat (report,s);
	sprintf (s, "Number of nodes: %d\n",numnodes);
	strcat (report,s);
	sprintf (s, "Number of links: %d\n",numlinks);
	strcat (report,s);
	sprintf (s, "Number of sides: %d\n",numlinks+numbranches);
	strcat (report,s);
	sprintf (s, "Number of holes: %d\n",numholes);
	strcat (report,s);
	sprintf (s, "Number separate loops: %d\n",numloops);
	strcat (report,s);
	sprintf (s, "Euler number: %d\n",eulernum);
	strcat (report,s);
	sprintf (s, "Average branch length: %7.1f\n",tbranchlen);
	strcat (report,s);
	sprintf (s, "Average link length: %7.1f\n",tlinklen);
	strcat (report,s);
	sprintf (s, "Average side length: %7.1f\n",(tlinklen*numlinks+tbranchlen*numbranches)/(numlinks+numbranches));
	strcat (report,s);
	sprintf (s, "Average loop length: %7.1f\n",looplen);
	strcat (report,s);
	
	/* If this is called from a macro, fill the user array */
	
	if (macrorun)
	{
		i=0;
		mac_user[i++]=numelems;					/* 0: Num elements */
		mac_user[i++]=dust;						/* 1: num isolated dots */
		mac_user[i++]=numends;					/* 2: Num ends */
		mac_user[i++]=numbranches;				/* 3: Num  branches */
		mac_user[i++]=numnodes;					/* 4: Num nodes */
		mac_user[i++]=numlinks;					/* 5: Num links */
		mac_user[i++]=numlinks+numbranches;		/* 6: Num sides */
		mac_user[i++]=numholes; 				/* 7: Num holes */
		mac_user[i++]=numloops;					/* 8: Num separate loops */
		mac_user[i++]=eulernum;					/* 9: Euler */
		mac_user[i++]=branchlen;				/* 10: Average branch length */
		mac_user[i++]=linklen;					/* 11: Average link length */
		mac_user[i++]=(linklen*numlinks+branchlen*numbranches)/(numlinks+numbranches);	/* 12: Average side length */
		mac_user[i++]=looplen;					/* Average loop length */
	}

	/* Return the new image, because the color coded topological points are more fun to see */
	handback (img, &tempimg);
	make_minmax (img);

}


void skeletonize_test (image_type* img, int slice)
{
image_type dest;
long cnt,i;
int x,y;
unsigned char* p;
unsigned char* p1;
unsigned char* p2;

	g_assert (img->imgtype==B_CHAR);

	if (allocate_image (&dest, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p2 = (unsigned char*) calloc (img->xmax*img->ymax,sizeof(unsigned char));
	if (!p2) fatal_memerr();
	
	skel_thresh (img, &dest, slice, 1);
	p = dest.data;
	p1= img->data;

	do
	{
		cnt = skel_iterate (&dest, p2); i++;
		dp (1, "Skeletonize (test): points deleted: %li\n",cnt);

		/* Reduce gray value of deleted points to indicate iteration */
	
		for (y=0; y<dest.ymax; y++)
			for (x=0; x<dest.xmax; x++)
			{
				if ((p1[x + img->xmax*y]) && (!p[x + dest.xmax*y])) 
					p1[x+img->xmax*y] --;
			}
	}
	while (cnt>0);
		
	free (p2);
	freebuf (&dest);
}


/************************************************
/
/	GUI drivers for skeletonize function 
/ 	based on threshold dialog
/
************************************************/


void skel_driver (float threshold, int prunelen, int rpt)
{
char report[16384];

	undo_nexus (&mainimg);
	
	skeletonize (&mainimg, cur_slice, threshold);
	make_minmax (&mainimg);
	if (prunelen) skel_prune (&mainimg, prunelen);
	if (rpt)
	{
		skel_statistics (&mainimg, report);
		ShowLongMessage (report, "Skeleton Statistics");
	}

	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
	
}

GtkAdjustment *skel_t_adj;
GtkAdjustment *skel_prune_adj;
int skel_report;


void on_skel_ok_clicked ( GtkButton * button, gpointer data )
{
int threshold, prunelen;

	threshold = skel_t_adj->value;
	prunelen = skel_prune_adj->value;

	skel_driver (threshold,prunelen, skel_report);
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_skel_cancel_clicked ( GtkButton * button, gpointer data )
{
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}

void on_skel_button_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	skel_report = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Skeleton report output set to %d\n", skel_report);

}



void create_skel_wnd (image_type* img)
{
  GtkWidget *skel_wnd, *skel_vbox, *skel_table1;
  GtkWidget *skel_label2;
  GtkWidget *skel_label3;
  GtkWidget *skel_label4;
  GtkWidget *skel_label5;
  GtkWidget *skel_okbutton;
  GtkWidget *skel_cancelbutton;
  GtkWidget *skel_t;
  GtkWidget *skel_prune;
  GtkWidget *hsep;
  GtkWidget *skel_button;

  
  float min,max;
   
	if (img->imgtype == B_FLOAT) max = img->maxval.fmax; else max = img->maxval.imax;
	if (img->imgtype == B_FLOAT) min = img->minval.fmin; else min = img->minval.imin;

	skel_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (skel_wnd), "skeles_wnd", skel_wnd);
	gtk_window_set_title (GTK_WINDOW (skel_wnd), "Skeletonize");
	gtk_window_set_modal (GTK_WINDOW (skel_wnd), TRUE);
	skel_vbox = vbox_in_window (skel_wnd, "skel_vbox1");
	gtk_container_set_border_width (GTK_CONTAINER (skel_vbox), 5);

	skel_table1 = MakeTable (skel_wnd, skel_vbox, "skel_table1", 2, 5);
	
	skel_label2 = MakeLabel (skel_wnd, skel_table1, "Threshold", "skel_label2", 0,0);
	skel_label3 = MakeLabel (skel_wnd, skel_table1, "Prune", "skel_label3", 0,1);
	skel_label4 = MakeLabel (skel_wnd, skel_table1, "Generate report", "skel_label4", 0,2);

	skel_t_adj = (GtkAdjustment*)gtk_adjustment_new (min+1, min, max, 1, 10, 0);
	skel_t = MakeSpinbutton (GTK_ADJUSTMENT(skel_t_adj), skel_wnd, skel_table1, "skel_t", 1,0);
	skel_prune_adj = (GtkAdjustment*)gtk_adjustment_new (0, 0, 100, 1, 10, 0);
	skel_prune = MakeSpinbutton (GTK_ADJUSTMENT(skel_prune_adj), skel_wnd, skel_table1, "skel_t", 1,1);
	skel_button = Checkbutton_in_table (skel_wnd,skel_table1," ","skel_tabbtn",1,2,NULL);
	gtk_signal_connect (GTK_OBJECT (skel_button), "toggled",
                    	GTK_SIGNAL_FUNC (on_skel_button_clicked), skel_button);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (skel_button), TRUE);
	skel_report=1;

	hsep = MakeHsep (skel_wnd, skel_table1, "hseparator4", 0,3);
	hsep = MakeHsep (skel_wnd, skel_table1, "hseparator5", 1,3);


	skel_okbutton = Button_in_table (skel_wnd, skel_table1, "OK", "skel_okbutton", 0,4, NULL);
	skel_cancelbutton = Button_in_table (skel_wnd, skel_table1, "Cancel", "skel_canbutton", 1,4, NULL);

	gtk_signal_connect (GTK_OBJECT (skel_okbutton), "clicked", 
                    	GTK_SIGNAL_FUNC (on_skel_ok_clicked), skel_wnd);
	gtk_signal_connect (GTK_OBJECT (skel_cancelbutton), "clicked", 
                    	GTK_SIGNAL_FUNC (on_skel_cancel_clicked), skel_wnd);
	gtk_signal_connect (GTK_OBJECT (skel_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_skel_cancel_clicked), skel_wnd);


	gtk_widget_show (skel_wnd);
}



void skel_main ()
{
	create_skel_wnd (&mainimg);
}



/******************************************************************

 #   #    *##    ####    ####    #   #            *##    ####   
 ## ##   #   #   #   #   #   #   #   #           #   #   #   #  
 # # #   #   #   #   #   #   #   #   #           #   #   #   #  
 # # #   #   #   ####    ####    #####           #   #   ####   
 #   #   #   #   # #     #       #   #           #   #   #      
 #   #   #   #   #  #    #       #   #           #   #   #      
 #   #    *##    #   #   #       #   #            *##    #      
 
/******************************************************************/



/*************************************************************

	Morphological operators: Erode, Dilate.
	The UI here provides two steps in succession,
	allowing to combine erode with dilate (=open) and
	dilate with erode (=close).
	Other morphological operators of the same class
	include median fitering and my own special variants:
	Soft erode: Set to average of 8 neighbors if this is smaller than the value
	Soft dilate: Same, but if average is greater than value
	Patch: Set to average if 5 or more pixels are nonzero
	
	Finally, an option allows the resulting image to be
	X-ored with the original image if it is not a float image
	
***************************************************************/




/* morph_value returns one element for the new center
   of a 3x3 neighborhood. We need two different functions
   to handle float and non-float. We can pass n=9 for
   an 8-neighborhood and n=5 for the 4-neighborhood.
	The exception is the CWMF which requires 13 elements, always. */


float morph_val_float (MORPHOP what, float* m, int n, float origval)
{
float mval, f, b;
int i, j, k;

	mval=0; 
	if (what==M_CWMF) n=13;			/* Fixed number of elements for center-weighted median */
	switch (what)
   	{
		case M_BYPASS:
		{
			mval = origval;
			break;
		}
		case M_ERODE:  /* Minimum value of 8 neighbors */
		{
	    	mval=m[0]; for (i=1; i<n; i++) { if (mval>m[i]) mval=m[i]; }
	    	break;
	 	}
      	case M_SOFTERODE:  /* Average of 8 neighbors if eroding */
        {
	    	mval=m[0]; for (i=1; i<n; i++) { mval+=m[i]; }
	    	mval /= n;
	    	if (mval > origval) mval=origval; 
	    	break;
	 	}
      	case M_DILATE:   /* Maximum of 8 neighbors */
        {
	    	mval=m[0]; for (i=1; i<n; i++) { if (mval<m[i]) mval=m[i]; }
	    	break;
	 	}
      	case M_SOFTDILATE:  /* Average of 8 neighbors if dilating */
        {
	    	mval=m[0]; for (i=1; i<n; i++) { mval+=m[i]; }
	    	mval /= n;
	    	if (origval > mval) mval=origval; 
	    	break;
	 	}
      	case M_RANK:    /* Median of the neighbor values */
        {
			/* Quick-n-dirty sort algorithm, inefficient but short - and replaced by combsort 
        	for (i=0; i<n-1; i++) 
        	{   
	       		f=m[i]; k=i;
               	for (j=i+1; j<n; j++) { if (f>m[j]) { k=j; f=m[k]; }; };
               	if (i != k) { b=m[i]; m[i]=m[k]; m[k]=b; };
        	}
			*/
			combsort_f (m, NULL, n);
            mval = m[n/2];
			break;
	 	}
      	case M_CWMF:    /* Center-weighted median of the neighbor values, MUST be 8-neighborhood */
        {
			combsort_f (m, NULL, n);
            mval = m[n/2];
			break;
	 	}
      	case M_PATCH:   	/* Average of the 8 neighbors if more than 1/2 of */
        {             		/* the environment is > 0 */
	    	mval=origval;
	    	if (origval==0)
	    	{
	       		k=0; mval=0; for (i=0; i<n; i++) { mval+=m[i]; if (m[i]!=0) k++; }
		    	mval /= n;
				if (k<=n/2) mval=origval;
	    	}
			break;
	 	}
      	case M_BOUNDERODE: 	/* Erode (set zero) only if boundary point */
        {
	    	mval=origval;
	    	if (origval!=0)
	    	{
				for (i=0; i<n; i++) { if (m[i]==0) mval=0; }
	    	}
			break;
	 	}
   	}
   	return mval;
}



int morph_val_int (MORPHOP what, int* m, int n, int origval)
{
int mval, f, b;
int i, j, k;

	mval=0; 
	if (what==M_CWMF) n=13;			/* Fixed number of elements for center-weighted median */
	switch (what)
   	{
		case M_BYPASS:
		{
			mval = origval;
			break;
		}
		case M_ERODE:  /* Minimum value of 8 neighbors */
		{
	    	mval=m[0]; for (i=1; i<n; i++) { if (mval>m[i]) mval=m[i]; }
	    	break;
	 	}
      	case M_SOFTERODE:  /* Average of 8 neighbors if eroding */
        {
	    	mval=m[0]; for (i=1; i<n; i++) { mval+=m[i]; }
	    	mval = floor (0.5+(float)mval/(float)n);
	    	if (mval > origval) mval=origval; 
	    	break;
	 	}
      	case M_DILATE:   /* Maximum of 8 neighbors */
        {
	    	mval=m[0]; for (i=1; i<n; i++) { if (mval<m[i]) mval=m[i]; }
	    	break;
	 	}
      	case M_SOFTDILATE:  /* Average of 8 neighbors if dilating */
        {
	    	mval=m[0]; for (i=1; i<n; i++) { mval+=m[i]; }
	    	mval = floor (0.5+(float)mval/(float)n);
	    	if (origval > mval) mval=origval; 
	    	break;
	 	}
      	case M_RANK:    /* Median of the neighbor values */
        {
	
			/* Quick-n-dirty sort algorithm - replaced by combsort
        	for (i=0; i<n-1; i++) 
        	{   
	       		f=m[i]; k=i;
               	for (j=i+1; j<n; j++) { if (f>m[j]) { k=j; f=m[k]; }; };
               	if (i != k) { b=m[i]; m[i]=m[k]; m[k]=b; };
        	}
			*/
			combsort_i (m, NULL, n);
            mval = m[n/2];
			break;
	 	}
      	case M_CWMF:    /* Center-weighted median of the neighbor values, MUST be 8-neighborhood */
        {
			combsort_i (m, NULL, n);
            mval = m[n/2];
			break;
	 	}
      	case M_PATCH:   	/* Average of the 8 neighbors if more than 1/2 of */
        {             		/* the environment is > 0 */
	    	mval=origval;
	    	if (origval==0)
	    	{
	       		k=0; mval=0; 
				for (i=0; i<n; i++) { mval+=m[i]; if (m[i]!=0) k++; }
		    	if (k<=n/2) 
					mval=origval;
				else
					mval = floor (0.5+(float)mval/(float)k);

	    	}
			break;
	 	}
      	case M_BOUNDERODE: 	/* Erode (set zero) only if boundary point */
        {
	    	mval=origval;
	    	if (origval!=0)
	    	{
				for (i=0; i<n; i++) { if (m[i]==0) mval=0; }
	    	}
			break;
	 	}
   	}
   	return mval;
}


/* Morphological operator acting on one slice of the image.
   With multiple iterations we alternate between 8- and
   4-neighborhood to better approximate a circular effect. The
   variable foureight determines this behavior (foureight==1, 8 neighb).
   
   What we do is to interlace the storage in the mi/mf array.
   By default, we fill it in the sequence NW-N-NE-W-C-E-SW-S-SE.
   A translation array xlat[] skews the order to N-W-C-E-S-NW-NE-SW-SE
   so that the first 5 elements represent the 4-neighborhood.
*/


void morph_img (image_type* img, MORPHOP what, int z, int foureight)
{
int x,y,i,j,x1,y1,n,neigh;
float mf[13];
int mi[13];				/* We only use 9 of them unless the CWMF is activated */
int xmax, ymax;
long idx;
/*      from:  NW  N NE  W  C  E SW  S SE 		note for CWMF the order is broken and we can't do 4-neighborhood.
        to:     N  W  C  E  S NW NE SW SE*/
int xlat[13] = { 5, 0, 6, 1, 2, 3, 7, 4, 8, 9, 10, 11, 12 };

	if (foureight) neigh=9; else neigh=5;
	dp (2, "Morphop: Operation %d on slice %d\n",what,z);

	xmax = img->xmax; ymax = img->ymax;
	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = (unsigned char*)img->data + z*xmax*ymax;
			unsigned char* p1 = (unsigned char*)calloc (xmax*ymax, sizeof (char));
			idx=0;
			for (y=0; y<ymax; y++)
				for (x=0; x<xmax; x++)
				{
					n=0;
					for (j=-1; j<=1; j++)
					{
						y1=y+j; if (y1<0) y1=0; else if (y1>=ymax) y1=ymax-1;
						for (i=-1; i<=1; i++)
						{
							x1=x+i; if (x1<0) x1=0; else if (x1>=xmax) x1=xmax-1;
							mi[xlat[n++]]=p[x1+xmax*y1];
						}
					}
					if (what==M_CWMF) for (j=1; j<4; j++) mi[xlat[n++]]=p[x+xmax*y];
					p1[x+xmax*y]=morph_val_int (what, mi, neigh, p[x+xmax*y]);
				}
			memcpy (p, p1, xmax*ymax*sizeof(char));
			free (p1);
			break;
		}
		case B_SHORT:
		{
			short* p = (short*)img->data + z*xmax*ymax;
			short* p1 = (short*)calloc (xmax*ymax, sizeof (short));
			idx=0;
			for (y=0; y<ymax; y++)
				for (x=0; x<xmax; x++)
				{
					n=0;
					for (j=-1; j<=1; j++)
					{
						y1=y+j; if (y1<0) y1=0; else if (y1>=ymax) y1=ymax-1;
						for (i=-1; i<=1; i++)
						{
							x1=x+i; if (x1<0) x1=0; else if (x1>=xmax) x1=xmax-1;
							mi[xlat[n++]]=p[x1+xmax*y1];
						}
					}
					if (what==M_CWMF) for (j=1; j<4; j++) mi[xlat[n++]]=p[x+xmax*y];
					p1[x+xmax*y]=morph_val_int (what, mi, neigh, p[x+xmax*y]);
				}
			memcpy (p, p1, xmax*ymax*sizeof(short));
			free (p1);
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data + z*xmax*ymax;
			float* p1 = (float*)calloc (xmax*ymax, sizeof (float));
			idx=0;
			for (y=0; y<ymax; y++)
				for (x=0; x<xmax; x++)
				{
					n=0;
					for (j=-1; j<=1; j++)
					{
						y1=y+j; if (y1<0) y1=0; else if (y1>=ymax) y1=ymax-1;
						for (i=-1; i<=1; i++)
						{
							x1=x+i; if (x1<0) x1=0; else if (x1>=xmax) x1=xmax-1;
							mf[xlat[n++]]=p[x1+xmax*y1];
						}
					}
					if (what==M_CWMF) for (j=1; j<4; j++) mf[xlat[n++]]=p[x+xmax*y];
					p1[x+xmax*y]=morph_val_float (what, mf, neigh, p[x+xmax*y]);
				}
			memcpy (p, p1, xmax*ymax*sizeof(float));
			free (p1);
			break;
		}
		case B_RGB:
		{
			rgbtriplet* p = img->data + z*xmax*ymax;
			rgbtriplet* p1 = (rgbtriplet*)calloc (xmax*ymax, sizeof (rgbtriplet));
			idx=0;
			for (y=0; y<ymax; y++)
				for (x=0; x<xmax; x++)
				{
					n=0;
					for (j=-1; j<=1; j++)
					{
						y1=y+j; if (y1<0) y1=0; else if (y1>=ymax) y1=ymax-1;
						for (i=-1; i<=1; i++)
						{
							x1=x+i; if (x1<0) x1=0; else if (x1>=xmax) x1=xmax-1;
							mi[xlat[n++]]=p[x1+xmax*y1].red;
						}
					}
					if (what==M_CWMF) for (j=1; j<4; j++) mi[xlat[n++]]=p[x+xmax*y].red;
					p1[x+xmax*y].red=morph_val_int (what, mi, neigh, p[x+xmax*y].red);

					n=0;
					for (j=-1; j<=1; j++)
					{
						y1=y+j; if (y1<0) y1=0; else if (y1>=ymax) y1=ymax-1;
						for (i=-1; i<=1; i++)
						{
							x1=x+i; if (x1<0) x1=0; else if (x1>=xmax) x1=xmax-1;
							mi[xlat[n++]]=p[x1+xmax*y1].green;
						}
					}
					if (what==M_CWMF) for (j=1; j<4; j++) mi[xlat[n++]]=p[x+xmax*y].green;
					p1[x+xmax*y].green=morph_val_int (what, mi, neigh, p[x+xmax*y].green);

					n=0;
					for (j=-1; j<=1; j++)
					{
						y1=y+j; if (y1<0) y1=0; else if (y1>=ymax) y1=ymax-1;
						for (i=-1; i<=1; i++)
						{
							x1=x+i; if (x1<0) x1=0; else if (x1>=xmax) x1=xmax-1;
							mi[xlat[n++]]=p[x1+xmax*y1].blue;
						}
					}
					if (what==M_CWMF) for (j=1; j<4; j++) mi[xlat[n++]]=p[x+xmax*y].blue;
					p1[x+xmax*y].blue=morph_val_int (what, mi, neigh, p[x+xmax*y].blue);

					idx++;
				}
			memcpy (p, p1, xmax*ymax*sizeof(rgbtriplet));
			free (p1);
			break;
		}
		
	}	/* End switch */

	img->uid = next_uid++;
}


/* perform a full morphology sequence: Level 1, then level 2 */

void morph_seq (image_type* img, MORPHOP fLevel1, int fL1depth, 
								 MORPHOP fLevel2, int fL2depth, int xorme, int z)
{
int x,y,i,j;
long idx;
image_type morphed;
float* pf; short* ps; unsigned char* pc; rgbtriplet* pr;
unsigned char c; short s;
char* morphtxt[9] = {"Bypass", "Patch", "Erode", "Soft Erode", "Median", "Soft Dilate", "Dilate", "Boundary erode", "CWMF"};

	pf=img->data; pc=img->data; ps=img->data; pr=img->data;
	idx = z*img->xmax*img->ymax;

	/* Create one slice of image buffer and copy the original image */

	if (allocate_image (&morphed,img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();

	switch (img->imgtype)
	{
		case B_CHAR:
			pc += idx;
			memcpy (morphed.data, pc, img->xmax*img->ymax);
			break;
		case B_SHORT:
			ps += idx;
			memcpy (morphed.data, ps, img->xmax*img->ymax*sizeof(short));
			break;
		case B_FLOAT:
			pf += idx;
			memcpy (morphed.data, pf, img->xmax*img->ymax*sizeof(float));
			break;
		case B_RGB:
			pr += idx;
			memcpy (morphed.data, pr, img->xmax*img->ymax*sizeof(rgbtriplet));
			break;
	}

	if (/* (fLevel1 != M_BYPASS) && */ (fL1depth>0))
	{
		dp (1, "Morphop, level 1: %s, %d iterations\n",morphtxt[fLevel1], fL1depth);
		for (i=0; i<fL1depth; i++)
			morph_img (&morphed, fLevel1, 0, !(i & 1));
	}
	
	if ((fLevel2 != M_BYPASS) && (fL2depth>0))
	{
		dp (1, "Morphop, level 2: %s, %d iterations\n",morphtxt[fLevel2], fL2depth);
		for (i=0; i<fL2depth; i++)
			morph_img (&morphed, fLevel2, 0, !(i & 1));
	}

	/* ... and copy it back into original image */
	
	if ((!xorme) || (img->imgtype == B_FLOAT))
	{
		switch (img->imgtype)
		{
			case B_CHAR:
				memcpy (pc, morphed.data, img->xmax*img->ymax);
				break;
			case B_SHORT:
				memcpy (ps, morphed.data, img->xmax*img->ymax*sizeof(short));
				break;
			case B_FLOAT:
				memcpy (pf, morphed.data, img->xmax*img->ymax*sizeof(float));
				break;
			case B_RGB:
				memcpy (pr, morphed.data, img->xmax*img->ymax*sizeof(rgbtriplet));
				break;
		}
	
	}
	else		/* More complicated if we XOR with the original */
	{
		
		dp (1, "XORing the result\n");
		switch (img->imgtype)
		{
			case B_CHAR:
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						c = ((unsigned char*)morphed.data)[x + y*img->xmax];
						pc[x + y*img->xmax] ^= c;
					}
				break;
			case B_SHORT:
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						s = ((short*)morphed.data)[x + y*img->xmax];
						ps[x + y*img->xmax] ^= s;
					}

				break;
			case B_RGB:
				j=0;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						c = ((unsigned char*)morphed.data)[j];
						pc[j++] ^= c;
						c = ((unsigned char*)morphed.data)[j];
						pc[j++] ^= c;
						c = ((unsigned char*)morphed.data)[j];
						pc[j++] ^= c;
					}

				break;
		}

	}

	freebuf (&morphed);
	
}



/*************************************************+

	And the GUI for the above operator ...
	
***************************************************/

MORPHOP fLevel1, fLevel2;
GtkAdjustment *l1_depth_adj, *l2_depth_adj;
GtkWidget* morph_wnd;
int morph_do3d, morph_do_xor;



void on_morph_cancel_clicked ( GtkButton * button, gpointer data )
{

	gtk_widget_destroy ( GTK_WIDGET ( morph_wnd ) );
}


void on_morph_ok_clicked ( GtkButton * button, gpointer data )
{
int z;

	undo_nexus (&mainimg);

	if (morph_do3d)
	{
		for (z=0; z<mainimg.zmax; z++)
			morph_seq (&mainimg, fLevel1, l1_depth_adj->value,
								 fLevel2, l2_depth_adj->value, morph_do_xor, z);
	}
	else
	{
		morph_seq (&mainimg, fLevel1, l1_depth_adj->value,
							 fLevel2, l2_depth_adj->value, morph_do_xor, cur_slice);
	}
	
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
	
	gtk_widget_destroy ( GTK_WIDGET ( morph_wnd ) );
}



void on_l1_menu_changed (GtkWidget *widget, gpointer user_data)
{
	fLevel1 = *(int*) user_data;		/* Note that the menuopt parameter must match the enum oreder of the filters */
}

void on_l2_menu_changed (GtkWidget *widget, gpointer user_data)
{
	fLevel2 = *(int*) user_data;
}


void on_morph_xor_clicked (GtkWidget *widget, gpointer user_data)
{
	morph_do_xor = !morph_do_xor;
}

void on_morph_3d_clicked (GtkWidget *widget, gpointer user_data)
{
	morph_do3d = !morph_do3d;
}


void create_morphwnd (image_type* img)
{
GtkWidget *vbox, *table1, *table2, *hbox;
GtkWidget *hsep1, *hsep2, *hsep3, *hsep4; 
GtkWidget *label1, *label2, *label3, *label4, *label5, *label6;
GtkWidget *menuitem, *l1_menu, *l1_menu_menu;
GtkWidget *l2_menu, *l2_menu_menu;
GtkWidget *l1_depth, *l2_depth;
GtkWidget *morph_okbutton, *morph_cancelbutton;
GtkWidget *xorbox, *do3dbox;

	morph_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (morph_wnd), "morph_wnd", morph_wnd);
	gtk_window_set_title (GTK_WINDOW (morph_wnd), "Morphological Operator");
	gtk_window_set_modal (GTK_WINDOW (morph_wnd), TRUE);

	vbox = vbox_in_window (morph_wnd, "morph_vbox1");
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);

	label1 = MakeLabel (morph_wnd,vbox,"Level 1", "label1",-1,-1);
	hsep1 = MakeHsep (morph_wnd, vbox, "morph_hseparator1", -1,-1);
	table1 = MakeTable (morph_wnd, vbox, "morph_tbl", 2,2);

	label2 = MakeLabel (morph_wnd,table1,"Operation", "label2",0,0);
	l1_menu = Make_Optionmenu (morph_wnd,table1,"l1_menu",1,0);
	l1_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Bypass", (gpointer) &menuopt0);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Patch", (gpointer) &menuopt1);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Erode", (gpointer) &menuopt2);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Soft Erode", (gpointer) &menuopt3);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Median", (gpointer) &menuopt4);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "CWMF (soft median)", (gpointer) &menuopt8);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Soft Dilate", (gpointer) &menuopt5);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Dilate", (gpointer) &menuopt6);
	menuitem = Option_menuitem (l1_menu_menu, GTK_SIGNAL_FUNC(on_l1_menu_changed), "Boundary Erode", (gpointer) &menuopt7);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (l1_menu), l1_menu_menu);

	label3 = MakeLabel (morph_wnd,table1,"Depth", "label3",0,1);
	l1_depth_adj = (GtkAdjustment*)gtk_adjustment_new (1,0,5,1,1,1);
	l1_depth = MakeSpinbutton (GTK_ADJUSTMENT(l1_depth_adj), morph_wnd,table1, "morph_l1", 1,1);

	hsep2 = MakeHsep (morph_wnd, vbox, "morph_hseparator2", -1,-1);
	label4 = MakeLabel (morph_wnd,vbox,"Level 2", "label4",-1,-1);
	hsep3 = MakeHsep (morph_wnd, vbox, "morph_hseparator3", -1,-1);

	table2 = MakeTable (morph_wnd, vbox, "morph_tbl", 2,2);
	label5 = MakeLabel (morph_wnd,table2,"Operation", "label5",0,0);
	l2_menu = Make_Optionmenu (morph_wnd,table2,"l2_menu",1,0);
	l2_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Bypass", (gpointer) &menuopt0);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Patch", (gpointer) &menuopt1);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Erode", (gpointer) &menuopt2);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Soft Erode", (gpointer) &menuopt3);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Median", (gpointer) &menuopt4);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "CWMF (soft median)", (gpointer) &menuopt8);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Soft Dilate", (gpointer) &menuopt5);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Dilate", (gpointer) &menuopt6);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_l2_menu_changed), "Boundary Erode", (gpointer) &menuopt7);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (l2_menu), l2_menu_menu);

	label6 = MakeLabel (morph_wnd,table2,"Depth", "label6",0,1);
	l2_depth_adj = (GtkAdjustment*)gtk_adjustment_new (1,0,5,1,1,1);
	l2_depth = MakeSpinbutton (GTK_ADJUSTMENT(l2_depth_adj), morph_wnd,table2, "morph_l2", 1,1);
	hsep4 = MakeHsep (morph_wnd, vbox, "morph_hseparator4", -1,-1);

	if (img->imgtype!=B_FLOAT)
		xorbox = Checkbutton_in_table (morph_wnd, vbox, "XOR result with source", "morph_xor",-1,-1,on_morph_xor_clicked);

	if (img->zmax>1)
		do3dbox = Checkbutton_in_table (morph_wnd, vbox, "Perform on all slices", "morph_3d",-1,-1,on_morph_3d_clicked);

	hbox = gtk_hbox_new (TRUE, 0);
	gtk_widget_ref (hbox);
	gtk_object_set_data_full (GTK_OBJECT (morph_wnd), "morph_hbox", hbox,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	morph_okbutton = Button_in_table (morph_wnd,hbox,"OK","morph_okbutton",-1,-1,on_morph_ok_clicked);
	morph_cancelbutton = Button_in_table (morph_wnd,hbox,"Cancel","morph_cancelbutton",-1,-1,on_morph_cancel_clicked);

	gtk_signal_connect (GTK_OBJECT (morph_wnd), "destroy",
                    	  GTK_SIGNAL_FUNC (on_morph_cancel_clicked), morph_wnd);

	gtk_widget_show (morph_wnd);
}


void morph_main ()
{

	morph_do3d = 0; morph_do_xor = 0;
	fLevel1 = M_BYPASS; fLevel2 = M_BYPASS;
	create_morphwnd (&mainimg);

}



